# This file is a part of the AnyBlok project
#
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Pierre Verkest <pverkest@anybox.fr>
#    Copyright (C) 2019 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.testing import sgdb_in
from anyblok.column import Integer as Int, String as Str
from anyblok.migration import (
    MigrationException, MigrationColumnTypePlugin, MigrationReport)
from anyblok.relationship import Many2Many
from contextlib import contextmanager
from sqlalchemy import (
    MetaData, Table, Column, Integer, String, Boolean, TEXT,
    CheckConstraint, ForeignKey, text
)
from sqlalchemy.dialects.mysql.types import TINYINT, DATETIME
from sqlalchemy.dialects.mssql.base import BIT
from anyblok import Declarations
from sqlalchemy.exc import IntegrityError
from anyblok.config import Configuration
from .conftest import init_registry, drop_database, create_database
from anyblok.common import naming_convention
from mock import patch


@pytest.fixture(scope="module")
def clean_db(request, configuration_loaded):

    def clean():
        url = Configuration.get('get_url')()
        drop_database(url)
        db_template_name = Configuration.get('db_template_name', None)
        create_database(url, template=db_template_name)

    request.addfinalizer(clean)


@contextmanager
def cnx(registry):
    cnx = registry.migration.conn
    try:
        yield cnx
    except Exception:
        cnx.execute(text("rollback"))
        raise


def add_in_registry():
    register = Declarations.register
    Model = Declarations.Model

    @register(Model)
    class Test:
        integer = Int(primary_key=True)
        other = Str()

    @register(Model)
    class TestUnique:
        integer = Int(primary_key=True)
        other = Str(unique=True)

    @register(Model)
    class TestIndex:
        integer = Int(primary_key=True)
        other = Str(index=True)

    if not sgdb_in(['MySQL', 'MariaDB', 'MsSQL']):
        @register(Model)
        class TestCheck:
            integer = Int(primary_key=True)

            @classmethod
            def define_table_args(cls):
                table_args = super(TestCheck, cls).define_table_args()
                return table_args + (
                    CheckConstraint('integer > 0', name='test'),)

        @register(Model)
        class TestCheckLongConstraintName:
            integer = Int(primary_key=True)

            @classmethod
            def define_table_args(cls):
                table_args = super(TestCheckLongConstraintName,
                                   cls).define_table_args()
                return table_args + (
                    CheckConstraint('integer > 0', name=(
                        'long_long_long_long_long_long_long_long_long_long_'
                        'long_long_long_long_long_long_long_long_test')),)

    @register(Model)
    class TestFKTarget:
        integer = Int(primary_key=True)

    @register(Model)
    class TestFK:
        integer = Int(primary_key=True)
        other = Int(
            foreign_key=Model.TestFKTarget.use('integer'))

    @register(Model)
    class TestFK2:
        integer = Int(primary_key=True)
        other = Int(
            foreign_key=Model.TestFKTarget.use('integer').options(
                ondelete='cascade'))

    @register(Model)
    class TestM2M1:
        idmodel1 = Int(primary_key=True)

    @register(Model)
    class TestM2M2:
        idmodel2 = Int(primary_key=True)
        rel_m2m = Many2Many(label="Rel", model=Model.TestM2M1,
                            join_table='reltable',
                            remote_columns='idmodel1',
                            m2m_remote_columns='idmodel1',
                            local_columns='idmodel2',
                            m2m_local_columns='idmodel2',
                            many2many='rel_m2m_inv')


@pytest.fixture()
def registry(request, clean_db, bloks_loaded):
    registry = init_registry(add_in_registry)

    def rollback():
        for table in ('reltable', 'test', 'test2', 'othername', 'testfk',
                      'testunique', 'reltab', 'testm2m1',
                      'testm2m2', 'testfk2', 'testfktarget',  'testcheck',
                      'testchecklongconstraintname', 'testindex'):
            try:
                registry.migration.table(table).drop()
            except MigrationException:
                pass

        registry.close()

    request.addfinalizer(rollback)
    return registry


class MockMigrationColumnTypePluginInteger2String(MigrationColumnTypePlugin):

    to_type = String
    from_type = Integer
    dialect = None


class MockMigrationColumnTypePluginInteger2StringMySQL(
    MigrationColumnTypePlugin
):

    to_type = String
    from_type = Integer
    dialect = ['MySQL']


class TestMigration:

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="Can't create empty table")
    def test_add_table(self, registry):
        registry.migration.table().add('test2')
        registry.migration.table('test2')

    def fill_test_table(self, registry):
        Test = registry.Test
        vals = [{'other': 'test %d' % x} for x in range(10)]
        Test.multi_insert(*vals)

    def test_add_column(self, registry):
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer))
        t.column('new_column')

    def test_add_column_in_filled_table(self, registry):
        self.fill_test_table(registry)
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer))
        t.column('new_column')

    def test_add_not_null_column(self, registry):
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer, nullable=False))
        t.column('new_column')

    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="Not rollback to savepoint")
    def test_add_not_null_column_in_filled_table(self, registry):
        self.fill_test_table(registry)
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer, nullable=False))
        t.column('new_column')

    def test_add_column_with_default_value(self, registry):
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer, default=100))
        t.column('new_column')

    def test_add_column_with_default_str_value(self, registry):
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer, default='100'))
        t.column('new_column')

    def test_add_column_in_filled_table_with_default_value(self, registry):
        self.fill_test_table(registry)
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer, default=100))
        t.column('new_column')
        res = [x for x in registry.execute(
            "select count(*) from test where new_column is null")][0][0]
        assert res == 0

    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="Can't change server default")
    def test_add_not_null_column_in_filled_table_with_default_value(
        self, registry
    ):
        self.fill_test_table(registry)
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer, nullable=False,
                       server_default="1"))
        t.column('new_column')

    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="Not rollback to savepoint")
    def test_add_unique_constraint_on_good_table(self, registry):
        self.fill_test_table(registry)
        t = registry.migration.table('test')
        t.unique('unique_constraint').add(t.column('other'))
        registry.Test.insert(other='One entry')
        with pytest.raises(IntegrityError):
            registry.Test.insert(other='One entry')

    @pytest.mark.skipif(sgdb_in(['MsSQL']), reason="#121")
    def test_add_unique_constraint_on_not_unique_column(self, registry):
        Test = registry.Test
        vals = [{'other': 'test'} for x in range(10)]
        Test.multi_insert(*vals)
        t = registry.migration.table('test')
        t.unique(None).add(t.column('other'))
        registry.Test.insert(other='One entry')
        registry.Test.insert(other='One entry')

    def test_drop_table(self, registry):
        registry.migration.table('test').drop()
        with pytest.raises(MigrationException):
            registry.migration.table('Test')

    def test_drop_column(self, registry):
        t = registry.migration.table('test')
        t.column('other').drop()
        with pytest.raises(MigrationException):
            t.column('other')

    def test_alter_table_name(self, registry):
        t = registry.migration.table('test')
        t.alter(name='othername')
        registry.migration.table('othername')
        with pytest.raises(MigrationException):
            registry.migration.table('test')

    def test_alter_column_name(self, registry):
        t = registry.migration.table('test')
        t.column('other').alter(name='othername')
        t.column('othername')
        with pytest.raises(MigrationException):
            t.column('other')

    def test_alter_column_nullable(self, registry):
        t = registry.migration.table('test')
        c = t.column('other').alter(nullable=False)
        assert not c.nullable

    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="Not rollback to savepoint")
    def test_alter_column_nullable_in_filled_table(self, registry):
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer))
        self.fill_test_table(registry)
        c = t.column('new_column').alter(nullable=False)
        # the column doesn't change of nullable to not lock the migration
        assert c.nullable

    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="Can't change server default")
    def test_alter_column_default(self, registry):
        t = registry.migration.table('test')
        c = t.column('other').alter(server_default='test')
        if sgdb_in(['PostgreSQL']):
            assert c.server_default == "'test'::character varying"
        else:
            assert str(c.server_default) == "'test'"

    def test_index(self, registry):
        t = registry.migration.table('test')
        t.index().add(t.column('integer'))
        t.index(t.column('integer')).drop()

    def test_alter_column_type(self, registry):
        t = registry.migration.table('test')
        c = t.column('other').alter(type_=TEXT)
        assert isinstance(c.type, TEXT)

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Can't drop primary key issue #92")
    def test_alter_column_primary_key(self, registry):
        t = registry.migration.table('test')
        t.primarykey().drop().add(t.column('other'))

    def test_alter_column_foreign_key(self, registry):
        t = registry.migration.table('test')
        t.foreign_key('my_fk').add(
            ['other'], [registry.System.Blok.name])
        t.foreign_key('my_fk').drop()

    def test_constraint_unique(self, registry):
        t = registry.migration.table('test')
        t.unique(name='test_unique_constraint').add(t.column('other'))
        t.unique(name='test_unique_constraint').drop()

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="Can't drop check constraint issue #93")
    def test_constraint_check(self, registry):
        t = registry.migration.table('test')
        Test = registry.Test
        t.check('test').add(Test.other != 'test')
        # particuliar case of check constraint
        t.check('anyblok_ck_test__test').drop()

    def test_detect_under_noautocommit_flag(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), nullable=False),
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.detect_changed()
        registry.migration.withoutautomigration = True
        with pytest.raises(MigrationException):
            registry.migration.detect_changed()

    def test_detect_column_added(self, registry):
        # Remove a column on the table force the detection to found new column
        # which is existing in metadata but not in table
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True)
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Add test.other")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not report.log_has("Add test.other")

    def test_detect_table_removed(self, registry):
        with cnx(registry) as conn:
            Table(
                'test2', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                Column('other2', String(64)),
            ).create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Drop Table test2")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert report.log_has("Drop Table test2")

    def test_detect_table_removed_with_reinit_column(self, registry):
        with cnx(registry) as conn:
            Table(
                'test2', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                Column('other2', String(64)),
            ).create(bind=conn)

        registry.migration.reinit_tables = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop Table test2")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Drop Table test2"))

    def test_detect_table_removed_with_reinit_all(self, registry):
        with cnx(registry) as conn:
            Table(
                'test2', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                Column('other2', String(64)),
            ).create(bind=conn)

        registry.migration.reinit_all = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop Table test2")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Drop Table test2"))

    def test_detect_column_removed(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                Column('other2', String(64)),
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Drop Column test.other2")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert report.log_has("Drop Column test.other2")

    def test_detect_column_removed_with_reinit_column(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                Column('other2', String(64)),
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.reinit_columns = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop Column test.other2")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Drop Column test.other2"))

    def test_detect_column_removed_with_reinit_all(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                Column('other2', String(64)),
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.reinit_all = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop Column test.other2")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Drop Column test.other2"))

    def test_detect_not_nullable_column_removed(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                Column('other2', String(64), nullable=False),
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert not(report.log_has("Drop Column test.other2"))
        assert report.log_has("Drop Column test.other2 (not null)")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert report.log_has("Drop Column test.other2")
        assert not(report.log_has("Drop Column test.other2 (not null)"))

    def test_detect_nullable(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), nullable=False),
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Alter test.other")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Alter test.other"))

    def test_detect_m2m_primary_key(self, registry):
        with cnx(registry) as conn:
            Table('reltable', registry.declarativebase.metadata,
                  autoload_with=conn).drop(bind=conn)

            meta = MetaData()
            meta._add_table('testm2m1', None, registry.TestM2M1.__table__)
            meta._add_table('testm2m2', None, registry.TestM2M2.__table__)

            Table(
                'reltable', meta,
                Column('idmodel1', Integer, ForeignKey('testm2m1.idmodel1')),
                Column('idmodel2', Integer, ForeignKey('testm2m2.idmodel2')),
            ).create(bind=conn)

        with pytest.raises(MigrationException):
            registry.migration.detect_changed()

    def test_detect_server_default(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), server_default='9.99'),
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Alter test.other")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Alter test.other"))

    def test_detect_add_index_constraint(self, registry):
        with cnx(registry) as conn:
            registry.TestIndex.__table__.drop(bind=conn)
            registry.TestIndex.__table__ = Table(
                'testindex', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
            )
            registry.TestIndex.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Add index constraint on testindex (other)")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Add index constraint on testindex (other)"))

    def test_detect_add_column_with_index_constraint(self, registry):
        with cnx(registry) as conn:
            registry.TestIndex.__table__.drop(bind=conn)
            registry.TestIndex.__table__ = Table(
                'testindex', MetaData(),
                Column('integer', Integer, primary_key=True),
            )
            registry.TestIndex.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Add testindex.other")
        assert report.log_has("Add index constraint on testindex (other)")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Add testindex.other"))
        assert not(report.log_has(
            "Add index constraint on testindex (other)"))

    def test_detect_drop_index(self, registry):
        with cnx(registry) as conn:
            conn.execute(text("CREATE INDEX other_idx ON test (other);"))
        report = registry.migration.detect_changed()
        assert report.log_has("Drop index other_idx on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert report.log_has("Drop index other_idx on test")

    def test_detect_drop_anyblok_index(self, registry):
        with cnx(registry) as conn:
            conn.execute(
                text("CREATE INDEX anyblok_ix_test__other ON test (other);"))
        report = registry.migration.detect_changed()
        assert report.log_has("Drop index anyblok_ix_test__other on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(
            report.log_has("Drop index anyblok_ix_test__other on test"))

    def test_detect_drop_index_with_reinit_indexes(self, registry):
        with cnx(registry) as conn:
            conn.execute(text("CREATE INDEX other_idx ON test (other);"))
        registry.migration.reinit_indexes = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop index other_idx on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Drop index other_idx on test"))

    def test_detect_drop_index_with_reinit_all(self, registry):
        with cnx(registry) as conn:
            conn.execute(text("CREATE INDEX other_idx ON test (other);"))
        registry.migration.reinit_all = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop index other_idx on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Drop index other_idx on test"))

    def test_detect_type(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', Integer),
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Alter test.other")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Alter test.other"))

    def test_detect_primary_key(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, nullable=False),
                Column('other', String(64), primary_key=True),
            )
            registry.Test.__table__.create(bind=conn)

        with pytest.raises(MigrationException):
            registry.migration.detect_changed()

    def test_detect_add_foreign_key(self, registry):
        with cnx(registry) as conn:
            registry.TestFK.__table__.drop(bind=conn)
            registry.TestFK.__table__ = Table(
                'testfk', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', Integer),
            )
            registry.TestFK.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has(
            "Add Foreign keys on (testfk.other) => (testfktarget.integer)")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Add Foreign keys on (testfk.other) => (testfktarget.integer)"))

    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="MsSQL does not change fk")
    def test_detect_foreign_key_options_changed(self, registry):
        with cnx(registry) as conn:
            registry.TestFK2.__table__.drop(bind=conn)
            meta = MetaData()
            meta._add_table(
                'testfktarget', None, registry.TestFKTarget.__table__)
            registry.TestFK2.__table__ = Table(
                'testfk2', meta,
                Column('integer', Integer, primary_key=True),
                Column('other', Integer, ForeignKey('testfktarget.integer')),
            )
            registry.TestFK2.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop Foreign keys on testfk2.other => testfktarget.integer")
        assert report.log_has(
            "Add Foreign keys on (testfk2.other) => (testfktarget.integer)")

    def test_detect_drop_foreign_key(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            meta = MetaData()
            meta._add_table(
                'system_blok', None, registry.System.Blok.__table__)
            registry.Test.__table__ = Table(
                'test', meta,
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), ForeignKey('system_blok.name')),
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop Foreign keys on test.other => system_blok.name")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop Foreign keys on test.other => system_blok.name")

    def test_detect_drop_anyblok_foreign_key(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            meta = MetaData(naming_convention=naming_convention)
            meta._add_table(
                'system_blok', None, registry.System.Blok.__table__)
            registry.Test.__table__ = Table(
                'test', meta,
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), ForeignKey('system_blok.name')),
            )
            registry.Test.__table__.create(bind=conn)
            # anyblok_fk_test__other_on_system_blok__name

        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop Foreign keys on test.other => system_blok.name")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Drop Foreign keys on test.other => system_blok.name"))

    def test_detect_drop_foreign_key_with_reinit_constraint(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            meta = MetaData()
            meta._add_table(
                'system_blok', None, registry.System.Blok.__table__)
            registry.Test.__table__ = Table(
                'test', meta,
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), ForeignKey('system_blok.name')),
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.reinit_constraints = True
        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop Foreign keys on test.other => system_blok.name")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Drop Foreign keys on test.other => system_blok.name"))

    def test_detect_drop_foreign_key_with_reinit_all(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            meta = MetaData()
            meta._add_table(
                'system_blok', None, registry.System.Blok.__table__)
            registry.Test.__table__ = Table(
                'test', meta,
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), ForeignKey('system_blok.name')),
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.reinit_all = True
        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop Foreign keys on test.other => system_blok.name")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Drop Foreign keys on test.other => system_blok.name"))

    def test_detect_drop_column_with_foreign_key(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            meta = MetaData(naming_convention=naming_convention)
            meta._add_table(
                'system_blok', None, registry.System.Blok.__table__)
            registry.Test.__table__ = Table(
                'test', meta,
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                Column('other2', String(64), ForeignKey('system_blok.name')),
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop Foreign keys on test.other2 => system_blok.name")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Drop Foreign keys on test.other2 => system_blok.name"))

    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="MsSQL does not add unique #121")
    def test_detect_add_unique_constraint(self, registry):
        with cnx(registry) as conn:
            registry.TestUnique.__table__.drop(bind=conn)
            registry.TestUnique.__table__ = Table(
                'testunique', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
            )
            registry.TestUnique.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Add unique constraint on testunique (other)")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Add unique constraint on testunique (other)"))

    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="MsSQL does not add unique #121")
    def test_detect_add_column_with_unique_constraint(self, registry):
        with cnx(registry) as conn:
            registry.TestUnique.__table__.drop(bind=conn)
            registry.TestUnique.__table__ = Table(
                'testunique', MetaData(),
                Column('integer', Integer, primary_key=True),
            )
            registry.TestUnique.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Add unique constraint on testunique (other)")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Add unique constraint on testunique (other)"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="MySQL transform unique constraint on index")
    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="MsSQL does not drop unique #121")
    def test_detect_drop_unique_constraint(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), unique=True),
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Drop constraint test_other_key on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert report.log_has("Drop constraint test_other_key on test")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="MySQL transform unique constraint on index")
    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="MsSQL does not drop unique #121")
    def test_detect_drop_unique_anyblok_constraint(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(naming_convention=naming_convention),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), unique=True),
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Drop constraint anyblok_uq_test__other on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(
            report.log_has("Drop constraint anyblok_uq_test__other on test"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="MySQL transform unique constraint on index")
    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="MsSQL does not drop unique #121")
    def test_detect_drop_unique_constraint_with_reinit_constraints(
        self, registry
    ):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), unique=True),
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.reinit_constraints = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop constraint test_other_key on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(
            report.log_has("Drop constraint test_other_key on test"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="MySQL transform unique constraint on index")
    @pytest.mark.skipif(sgdb_in(['MsSQL']),
                        reason="MsSQL does not drop unique #121")
    def test_detect_drop_unique_constraint_with_reinit_all(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64), unique=True),
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.reinit_all = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop constraint test_other_key on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(
            report.log_has("Drop constraint test_other_key on test"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="No CheckConstraint works #90")
    def test_no_detect_drop_and_add_check_constraint_with_long_name(
        self, registry
    ):
        report = registry.migration.detect_changed()
        assert not(
            report.log_has(
                "Drop check constraint anyblok_ck_testchecklongconstraintname"
                "__long_long_long_long_lon on testchecklongconstraintname"
            )
        )
        assert not(
            report.log_has(
                "Add check constraint anyblok_ck_testchecklongconstraintname__"
                "long_long_long_long_long_long_long_long_long_long_long_long_"
                "long_long_long_long_long_long_test on "
                "testchecklongconstraintname"
            )
        )

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="No CheckConstraint works #90")
    def test_detect_add_check_constraint(self, registry):
        with cnx(registry) as conn:
            registry.TestCheck.__table__.drop(bind=conn)
            registry.TestCheck.__table__ = Table(
                'testcheck', MetaData(naming_convention=naming_convention),
                Column('integer', Integer, primary_key=True),
            )
            registry.TestCheck.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has(
            "Add check constraint anyblok_ck_testcheck__test on testcheck")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Add check constraint anyblok_ck_testcheck__test on testcheck"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="No CheckConstraint works #90")
    def test_detect_drop_check_constraint(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                CheckConstraint("other != 'test'", name='ck_other')
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has("Drop check constraint ck_other on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert report.log_has("Drop check constraint ck_other on test")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="No CheckConstraint works #90")
    def test_detect_drop_check_anyblok_constraint(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(naming_convention=naming_convention),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                CheckConstraint("other != 'test'", name='check')
            )
            registry.Test.__table__.create(bind=conn)

        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop check constraint anyblok_ck_test__check on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Drop check constraint anyblok_ck_test__check on test"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="No CheckConstraint works #90")
    def test_detect_drop_check_constraint_with_reinit_constraint(
        self, registry
    ):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                CheckConstraint("other != 'test'", name='ck_other')
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.reinit_constraints = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop check constraint ck_other on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(
            report.log_has("Drop check constraint ck_other on test"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason="No CheckConstraint works #90")
    def test_detect_drop_check_constraint_with_reinit_all(self, registry):
        with cnx(registry) as conn:
            registry.Test.__table__.drop(bind=conn)
            registry.Test.__table__ = Table(
                'test', MetaData(),
                Column('integer', Integer, primary_key=True),
                Column('other', String(64)),
                CheckConstraint("other != 'test'", name='ck_other')
            )
            registry.Test.__table__.create(bind=conn)

        registry.migration.reinit_all = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop check constraint ck_other on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(
            report.log_has("Drop check constraint ck_other on test"))

    @pytest.mark.skipif(
        sgdb_in(['MySQL', 'MariaDB']),
        reason=(
            "Modification of schema create an implicite commit "
            "with MySQL and MariaDB, save point is by passed for them"
        )
    )
    def test_savepoint(self, registry):
        Test = registry.Test
        self.fill_test_table(registry)
        registry.migration.savepoint('test')
        self.fill_test_table(registry)
        assert Test.query().count() == 20
        registry.migration.rollback_savepoint('test')
        assert Test.query().count() == 10
        registry.migration.release_savepoint('test')


@pytest.fixture()
def registry_plugin(request, clean_db, bloks_loaded):
    registry = init_registry(add_in_registry)
    request.addfinalizer(registry.close)
    return registry


class TestMigrationPlugin:

    @pytest.mark.skipif(
        not sgdb_in(['MySQL', 'MariaDB']),
        reason='Plugin for MySQL only')
    def test_boolean_with_mysql(self, registry_plugin):
        report = MigrationReport(registry_plugin.migration, [])
        res = report.init_modify_type(
            [None, None, 'test', 'other', {}, TINYINT(), Boolean()])
        assert res is True

    @pytest.mark.skipif(
        not sgdb_in(['MySQL', 'MariaDB']),
        reason='Plugin for MySQL only')
    def test_datetime_with_mysql(self, registry_plugin):
        report = MigrationReport(registry_plugin.migration, [])
        res = report.init_modify_type(
            [None, None, 'test', 'other', {}, DATETIME(), DATETIME()])
        assert res is True

    @pytest.mark.skipif(
        not sgdb_in(['MsSQL']),
        reason='Plugin for MsSQL only')
    def test_boolean_with_mssql(self, registry_plugin):
        report = MigrationReport(registry_plugin.migration, [])
        res = report.init_modify_type(
            [None, None, 'test', 'other', {}, BIT(), Boolean()])
        assert res is True

    @pytest.mark.skipif(
        not sgdb_in(['PostgreSQL']),
        reason='Plugin for MsSQL only')
    def test_boolean_with_postgres(self, registry_plugin):
        report = MigrationReport(registry_plugin.migration, [])
        res = report.init_modify_type(
            [None, None, 'test', 'other', {}, Integer(), Boolean()])
        assert res is False

    def test_alter_column_type_with_plugin_1(self, registry_plugin):
        report = MigrationReport(registry_plugin.migration, [])
        report.plugins = [
            MockMigrationColumnTypePluginInteger2String,
        ]
        with patch(
            'anyblok.tests.test_migration.'
            'MockMigrationColumnTypePluginInteger2String.apply'
        ) as mockapply:
            report.apply_change_modify_type(
                [None, None, 'test', 'other', {}, Integer(), String()])
            mockapply.assert_called()

    def test_alter_column_type_with_plugin_2(self, registry_plugin):
        report = MigrationReport(registry_plugin.migration, [])
        report.plugins = [
            MockMigrationColumnTypePluginInteger2String,
        ]
        with patch(
            'anyblok.migration.MigrationColumn.alter'
        ) as mockapply:
            report.apply_change_modify_type(
                [None, None, 'test', 'other', {}, String(), Integer()])
            mockapply.assert_called()

    @pytest.mark.skipif(
        not sgdb_in(['MySQL', 'MariaDB']),
        reason='Plugin for MySQL only')
    def test_alter_column_type_with_plugin_3(self, registry_plugin):
        report = MigrationReport(registry_plugin.migration, [])
        report.plugins = [
            MockMigrationColumnTypePluginInteger2StringMySQL,
        ]
        with patch(
            'anyblok.tests.test_migration.'
            'MockMigrationColumnTypePluginInteger2StringMySQL.apply'
        ) as mockapply:
            report.apply_change_modify_type(
                [None, None, 'test', 'other', {}, Integer(), String()])
            mockapply.assert_called()

    @pytest.mark.skipif(
        sgdb_in(['MySQL', 'MariaDB']),
        reason='Plugin for MySQL only')
    def test_alter_column_type_with_plugin_4(self, registry_plugin):
        report = MigrationReport(registry_plugin.migration, [])
        report.plugins = [
            MockMigrationColumnTypePluginInteger2StringMySQL,
        ]
        with patch(
            'anyblok.migration.MigrationColumn.alter'
        ) as mockapply:
            report.apply_change_modify_type(
                [None, None, 'test', 'other', {}, Integer(), String()])
            mockapply.assert_called()



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Pierre Verkest <pverkest@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from sqlalchemy import Table, Column, ForeignKeyConstraint
from sqlalchemy.orm import (relationships, backref, relationship, base,
                            attributes)
from sqlalchemy.schema import Column as SA_Column
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy import exc as sa_exc, util
from sqlalchemy_utils.functions import get_class_by_table
from .field import Field, FieldException
from .mapper import ModelAdapter, ModelAttribute, ModelRepr, format_schema
from anyblok.common import anyblok_column_prefix
from sqlalchemy.ext.orderinglist import OrderingList, _unsugar_count_from
from types import FunctionType

from logging import getLogger


logger = getLogger(__name__)


class RelationshipProperty(relationships.RelationshipProperty):

    def __init__(self, *args, **kwargs):
        self.relationship_field = kwargs.pop('relationship_field')
        super(RelationshipProperty, self).__init__(*args, **kwargs)

    def _generate_backref(self):  # noqa
        """Interpret the 'backref' instruction to create a
        :func:`.relationship` complementary to this one."""

        if self.parent.non_primary:
            return  # pragma: no cover
        if self.backref is not None and not self.back_populates:
            if isinstance(self.backref, util.string_types):
                backref_key, kwargs = self.backref, {}  # pragma: no cover
            else:
                backref_key, kwargs = self.backref
            mapper = self.mapper.primary_mapper()

            check = set(mapper.iterate_to_root()).\
                union(mapper.self_and_descendants)
            for m in check:
                if m.has_property(backref_key):
                    raise sa_exc.ArgumentError(  # pragma: no cover
                        "Error creating backref "
                        "'%s' on relationship '%s': property of that "
                        "name exists on mapper '%s'" %
                        (backref_key, self, m))

            # determine primaryjoin/secondaryjoin for the
            # backref.  Use the one we had, so that
            # a custom join doesn't have to be specified in
            # both directions.
            if self.secondary is not None:
                # for many to many, just switch primaryjoin/
                # secondaryjoin.   use the annotated
                # pj/sj on the _join_condition.
                pj = kwargs.pop(  # pragma: no cover
                    'primaryjoin',
                    self._join_condition.secondaryjoin_minus_local)
                sj = kwargs.pop(  # pragma: no cover
                    'secondaryjoin',
                    self._join_condition.primaryjoin_minus_local)
            else:
                pj = kwargs.pop(
                    'primaryjoin',
                    self._join_condition.primaryjoin_reverse_remote)
                sj = kwargs.pop('secondaryjoin', None)
                if sj:
                    raise sa_exc.InvalidRequestError(  # pragma: no cover
                        "Can't assign 'secondaryjoin' on a backref "
                        "against a non-secondary relationship."
                    )

            foreign_keys = kwargs.pop('foreign_keys',
                                      self._user_defined_foreign_keys)
            parent = self.parent.primary_mapper()
            kwargs.setdefault('viewonly', self.viewonly)
            kwargs.setdefault('post_update', self.post_update)
            kwargs.setdefault('passive_updates', self.passive_updates)
            self.back_populates = backref_key
            _relationship = RelationshipProperty2(
                parent, self.secondary,
                pj, sj,
                foreign_keys=foreign_keys,
                back_populates=self.key,
                relationship_field=self.relationship_field,
                **kwargs)

            mapper._configure_property(backref_key, _relationship)

        if self.back_populates:
            self._add_reverse_property(self.back_populates)


def register_descriptor(class_, key, comparator=None,
                        parententity=None, doc=None, relationship_field=None):
    manager = base.manager_of_class(class_)
    descriptor = relationship_field.InstrumentedAttribute(
        class_, key, comparator=comparator, parententity=parententity,
        relationship_field=relationship_field)
    descriptor.__doc__ = doc
    manager.instrument_attribute(key, descriptor)
    return descriptor


class RelationshipProperty2(relationships.RelationshipProperty):

    def __init__(self, *args, **kwargs):
        self.relationship_field = kwargs.pop('relationship_field')
        super(RelationshipProperty2, self).__init__(*args, **kwargs)

    def instrument_class(self, mapper):
        register_descriptor(
            mapper.class_,
            self.key,
            comparator=self.comparator_factory(self, mapper),
            parententity=mapper,
            doc=self.doc,
            relationship_field=self.relationship_field,
        )


class RelationShipList:  # don't inherit list

    def append(self, x):
        res = super(RelationShipList, self).append(x)
        self.relationship_field_append_value(x)
        return res

    def extend(self, L):
        res = super(RelationShipList, self).extend(L)
        for el in L:
            self.relationship_field_append_value(el)

        return res

    def insert(self, i, x):
        res = super(RelationShipList, self).insert(i, x)
        self.relationship_field_append_value(x)
        return res

    def remove(self, x):
        self.relationship_field_remove_value(x)
        return super(RelationShipList, self).remove(x)

    def pop(self, *args, **kwargs):
        res = super(RelationShipList, self).pop(*args, **kwargs)
        self.relationship_field_remove_value(res)
        return res

    def clear(self):
        for x in self:
            self.relationship_field_remove_value(x)

        return super(RelationShipList, self).clear()


class RelationShip(Field):
    """ RelationShip class

    The RelationShip class is used to define the type of SQL field Declarations

    Add a new relation ship type::

        @Declarations.register(Declarations.RelationShip)
        class Many2one:
            pass

    the relationship column are forbidden because the model can be used on
    the model
    """

    def __init__(self, *args, **kwargs):
        self.forbid_instance(RelationShip)
        if 'model' in kwargs:
            self.model = ModelAdapter(kwargs.pop('model'))
        else:
            raise FieldException("model is required attribut")

        super(RelationShip, self).__init__(*args, **kwargs)

        if 'info' not in self.kwargs:
            self.kwargs['info'] = {}

        self.kwargs['info']['remote_model'] = self.model.model_name
        self.backref_properties = {}

    def autodoc_get_properties(self):
        res = super(RelationShip, self).autodoc_get_properties()
        res['model'] = self.model
        return res

    def apply_instrumentedlist(self, registry, namespace, fieldname):
        """ Add the InstrumentedList class to replace List class as result
        of the query

        :param registry: current registry
        """
        InstrumentedList = registry.InstrumentedList
        if self.kwargs.get('collection_class'):
            collection_class = self.kwargs['collection_class']
            if isinstance(collection_class, FunctionType):
                if getattr(
                    collection_class, 'is_an_anyblok_instrumented_list', False
                ) is True:
                    InstrumentedList = self.kwargs['collection_class'](registry)

        self.kwargs['collection_class'] = InstrumentedList
        self.backref_properties['collection_class'] = registry.InstrumentedList

    def define_backref_properties(self, registry, namespace, properties):
        """ Add in the backref_properties, new property for the backref

        :param registry: current registry
        :param namespace: name of the model
        :param properties: properties known of the model
        """
        pass

    def format_backref(self, registry, namespace, fieldname, properties):
        """ Create the real backref, with the backref string and the
        backref properties

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        :param properties: properties known of the model
        """
        _backref = self.kwargs.get('backref')
        if not _backref:
            return

        if isinstance(_backref, (list, tuple)):
            _backref, backref_properties = _backref
            if backref_properties.get('collection_class'):
                collection_class = backref_properties['collection_class']
                if isinstance(collection_class, FunctionType):
                    if getattr(
                        collection_class, 'is_an_anyblok_instrumented_list',
                        False
                    ) is True:
                        backref_properties[
                            'collection_class'] = collection_class(registry)

            self.backref_properties.update(backref_properties)

        self.define_backref_properties(registry, namespace, properties)

        if self.backref_properties:
            self.kwargs['backref'] = backref(_backref,
                                             **self.backref_properties)
            mapper = ModelAttribute(self.model.model_name, _backref)
            if not mapper.is_declared(registry):
                mapper.add_fake_relationship(registry, namespace, fieldname)

    def get_relationship_cls(self):
        return relationship

    def get_sqlalchemy_mapping(self, registry, namespace, fieldname,
                               properties):
        """ Return the instance of the real field

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        :param properties: properties known of the model
        :rtype: sqlalchemy relation ship instance
        """
        self.model.check_model(registry)
        self.format_label(fieldname)
        self.kwargs['info']['label'] = self.label
        self.kwargs['info']['rtype'] = self.__class__.__name__
        self.apply_instrumentedlist(registry, namespace, fieldname)
        self.format_backref(registry, namespace, fieldname, properties)
        return self.get_relationship_cls()(
            self.model.modelname(registry), **self.kwargs)

    def must_be_declared_as_attr(self):
        """ Return True, because it is a relationship """
        return True

    def init_expire_attributes(self, registry, namespace, fieldname):
        """Init dict of expiration properties

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        """
        if namespace not in registry.expire_attributes:
            registry.expire_attributes[namespace] = {}

        if fieldname not in registry.expire_attributes[namespace]:
            registry.expire_attributes[namespace][fieldname] = set()


class RelationShipListMany2One:

    def relationship_field_append_value(self, value):
        for model_field, rfield in self.relationship_fied.link_between_columns:
            self.relationship_fied.apply_value_to(
                value, model_field, getattr(value, self.fieldname), rfield)

    def relationship_field_remove_value(self, value):
        for model_field, rfield in self.relationship_fied.link_between_columns:
            setattr(value, anyblok_column_prefix + model_field, None)


class Many2One(RelationShip):
    """ Define a relationship attribute on the model

    ::

        @register(Model)
        class TheModel:

            relationship = Many2One(label="The relationship",
                                    model=Model.RemoteModel,
                                    remote_columns="The remote column",
                                    column_names="The column which have the "
                                                 "foreigh key",
                                    nullable=True,
                                    unique=False,
                                    index=False,
                                    primary_key=False,
                                    one2many="themodels")

    If the ``remote_columns`` are not define then, the system takes the primary
    key of the remote model

    If the column doesn't exist, the column will be created. Use the
    nullable option.
    If the name is not filled, the name is "'remote table'_'remote colum'"

    :param model: the remote model
    :param remote_columns: the column name on the remote model
    :param column_names: the column on the model which have the foreign key
    :param nullable: If the column_names is nullable
    :param unique: If True, add the unique constraint on the columns
    :param index: If True, add the index constraint on the columns
    :param primary_key: If True, add the primary_key=True on the columns
    :param one2many: create the one2many link with this many2one
    """
    use_hybrid_property = True

    def __init__(self, **kwargs):
        super(Many2One, self).__init__(**kwargs)

        self._remote_columns = None
        if 'remote_columns' in kwargs:
            self._remote_columns = self.kwargs.pop('remote_columns')
            if not isinstance(self._remote_columns, (list, tuple)):
                self._remote_columns = [self._remote_columns]

        self.primary_key = self.kwargs.pop('primary_key', False)
        self.kwargs['info']['primary_key'] = self.primary_key

        self.nullable = False if self.primary_key else True
        nullable = self.kwargs.pop('nullable', True)
        if not self.primary_key:
            self.nullable = nullable

        self.kwargs['info']['nullable'] = self.nullable

        self.unique = self.kwargs.pop('unique', False)
        self.kwargs['info']['unique'] = self.unique

        self.index = self.kwargs.pop('index', False)
        self.kwargs['info']['index'] = self.index

        if 'one2many' in kwargs:
            self.kwargs['backref'] = backref = self.kwargs.pop('one2many')
            self.kwargs['info']['remote_name'] = backref
            if isinstance(backref, (list, tuple)):
                self.kwargs['info']['remote_name'] = backref[0]
                self.backref_properties.update(**backref[1])

        self._column_names = None
        if 'column_names' in kwargs:
            self._column_names = self.kwargs.pop('column_names')
            if not isinstance(self._column_names, (list, tuple)):
                self._column_names = [self._column_names]

        self.foreign_key_options = self.kwargs.pop('foreign_key_options', {})
        self.cascade = self.kwargs.pop('cascade', 'save-update, merge')

    def autodoc_get_properties(self):
        res = super(Many2One, self).autodoc_get_properties()
        res['remote_columns'] = self._remote_columns
        res['column_names'] = self._column_names
        res['unique'] = self.unique
        res['index'] = self.index
        res['primary_key'] = self.primary_key
        return res

    autodoc_omit_property_values = Field.autodoc_omit_property_values.union((
        ('remote_columns', None),
        ('column_names', None),
        ('unique', False),
        ('primary_key', False),
    ))

    def get_remote_columns(self, registry):
        if self._remote_columns is None:
            return self.model.primary_keys(registry)

        return [ModelAttribute(self.model.model_name, x)
                for x in self._remote_columns]

    def get_columns_names(self, registry, namespace, fieldname, remote_columns):
        if self._column_names is None:
            model = ModelRepr(namespace)
            column_names = model.foreign_keys_for(
                registry, self.model.model_name)
            if not column_names:
                column_names = []
                for x in remote_columns:
                    cname = fieldname + '_' + x.attribute_name
                    column_names.append(ModelAttribute(namespace, cname))
        else:
            column_names = [ModelAttribute(namespace, x)
                            for x in self._column_names]

        return column_names

    def update_local_and_remote_columns_names(self, registry, namespace,
                                              fieldname):
        self.remote_columns = self.get_remote_columns(registry)
        self.kwargs['info']['remote_columns'] = [x.attribute_name
                                                 for x in self.remote_columns]
        self.column_names = self.get_columns_names(
            registry, namespace, fieldname, self.remote_columns)

    def get_property(self, registry, namespace, fieldname, properties):
        """Return the property of the field

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        :param properties: properties known to the model
        """
        res = super(Many2One, self).get_property(
            registry, namespace, fieldname, properties)
        # force the info value in hybrid_property because since SQLAlchemy
        # 1.1.* the info is not propagate
        res.info = self.kwargs['info']
        return res

    def add_expire_attributes(self, registry, namespace, fieldname, cname):
        self.init_expire_attributes(registry, namespace, cname)
        registry.expire_attributes[namespace][cname].add((fieldname,))
        if self.kwargs.get('backref'):
            backref = self.kwargs['backref']
            if isinstance(backref, (list, tuple)):
                backref = backref[0]

            registry.expire_attributes[namespace][cname].add(
                (fieldname, backref))

    def update_properties(self, registry, namespace, fieldname, properties):
        """ Create the column which has the foreign key if the column doesn't
        exist

        :param registry: the registry which load the relationship
        :param namespace: the name space of the model
        :param fieldname: fieldname of the relationship
        :param propertie: the properties known
        """
        add_fksc = False
        self.link_between_columns = []
        self.model.check_model(registry)
        self.update_local_and_remote_columns_names(
            registry, namespace, fieldname)
        if fieldname in [x.attribute_name for x in self.column_names]:
            raise FieldException("The column_names and the fieldname %r are "
                                 "the same, please choose another "
                                 "column_names" % fieldname)

        self.kwargs['info']['local_columns'] = [
            x.attribute_name for x in self.column_names]
        remote_types = {x.attribute_name: x.native_type(registry)
                        for x in self.remote_columns}
        remote_columns = {x.attribute_name: x
                          for x in self.remote_columns}
        for cname in self.column_names:
            if cname.is_declared(registry):
                del remote_types[cname.get_fk_column(registry)]

        col_names = []
        fk_names = []
        for cname in self.column_names:
            self.add_expire_attributes(registry, namespace, fieldname,
                                       cname.attribute_name)
            if not cname.is_declared(registry):
                rc, remote_type = self.get_column_information(
                    registry, cname, remote_types, fieldname)
                cname.add_fake_column(registry)
                foreign_key = remote_columns[rc].get_fk_name(registry)
                self.create_column(cname, remote_type, foreign_key, properties)
                add_fksc = True
                fk_name = remote_columns[rc]
            else:
                fk_name = cname.get_fk_mapper(registry)

            col_names.append(cname.attribute_name)
            fk_names.append(fk_name.get_fk_name(registry))
            self.link_between_columns.append((cname.attribute_name,
                                              fk_name.attribute_name))

        if namespace == self.model.model_name:
            self.kwargs['remote_side'] = [
                properties[anyblok_column_prefix + x.attribute_name]
                for x in self.remote_columns]

        if (len(self.column_names) > 1 or add_fksc) and col_names and fk_names:
            self.col_names = col_names
            self.fk_names = fk_names
            properties['add_in_table_args'].append(self)

    def remote_model_is_a_table(self, registry):
        if self.model.model_name not in registry.loaded_namespaces:
            return True  # by default

        return hasattr(registry.get(self.model.model_name), '__table__')

    def update_table_args(self, registry, Model):
        """Add foreign key constraint in table args"""
        if not self.remote_model_is_a_table(registry):
            return []  # pragma: no cover

        return [
            ForeignKeyConstraint(self.col_names, self.fk_names,
                                 **self.foreign_key_options)
        ]

    def get_column_information(self, registry, cname, remote_types, fieldname):
        if len(remote_types) == 1:
            rc = [x for x in remote_types][0]
            return rc, remote_types[rc]
        else:
            rc = cname.get_fk_column(registry)
            if rc is None:
                rc = cname.attribute_name[len(fieldname) + 1:]

            if rc in remote_types:
                return rc, remote_types[rc]
            else:
                cname.get_fk_column(registry)
                raise FieldException("Can not create the local "
                                     "column %r" % cname.attribute_name)

    def apply_instrumentedlist(self, registry, namespace, fieldname):
        """ Add the InstrumentedList class to replace List class as result
        of the query

        :param registry: current registry
        """
        properties = {
            'fieldname': fieldname, 'relationship_fied': self}

        collection_class = self.backref_properties.get('collection_class', None)
        if (
            collection_class
            and isinstance(collection_class, FunctionType)
            and getattr(
                collection_class, 'is_an_anyblok_instrumented_list', False
            ) is True
        ):
            InstrumentedList = collection_class(
                registry, RelationShipListMany2One, RelationShipList,
                **properties)
        else:
            InstrumentedList = type(
                'InstrumentedList', (RelationShipListMany2One, RelationShipList,
                                     registry.InstrumentedList), properties)

        self.backref_properties['collection_class'] = InstrumentedList

        cascade = self.cascade
        if self.foreign_key_options.get('ondelete') == 'cascade':
            cascade += ', delete'
        self.backref_properties['cascade'] = cascade

    def create_column(self, cname, remote_type, foreign_key, properties):

        def wrapper(cls):
            return SA_Column(
                cname.attribute_name,
                remote_type,
                nullable=self.nullable,
                unique=self.unique,
                index=self.index,
                primary_key=self.primary_key,
                info=dict(label=self.label, foreign_key=foreign_key))

        properties[(anyblok_column_prefix +
                    cname.attribute_name)] = declared_attr(wrapper)
        properties['loaded_columns'].append(cname.attribute_name)
        properties['hybrid_property_columns'].append(cname.attribute_name)

        fget = self.wrap_getter_column(cname.attribute_name)
        fset = super(Many2One, self).wrap_setter_column(cname.attribute_name)
        fexp = self.wrap_expr_column(cname.attribute_name)

        for func in (fget, fset, fexp):
            func.__name__ = cname.attribute_name

        hybrid = hybrid_property(fget)
        hybrid = hybrid.setter(fset)
        hybrid = hybrid.expression(fexp)
        properties[cname.attribute_name] = hybrid

    def get_sqlalchemy_mapping(self, registry, namespace, fieldname,
                               properties):
        """ Create the relationship

        :param registry: the registry which load the relationship
        :param namespace: the name space of the model
        :param fieldname: fieldname of the relationship
        :param propertie: the properties known
        :rtype: Many2One relationship
        """
        self.kwargs['foreign_keys'] = '[%s]' % ', '.join(
            [x.get_complete_name(registry) for x in self.column_names])
        if not self.remote_model_is_a_table(registry):
            self.kwargs['viewonly'] = True  # pragma: no cover
            # we used primaryjoin and let the userto defined the good one
            # The foreign key does not exist, we should fine a good way to
            # the the local and remote columns

        return super(Many2One, self).get_sqlalchemy_mapping(
            registry, namespace, fieldname, properties)

    def apply_value_to(self, model_self, model_field, remote_self,
                       remote_field):
        if remote_self:
            value = getattr(remote_self,
                            anyblok_column_prefix + remote_field)
        else:
            value = None

        setattr(model_self, anyblok_column_prefix + model_field, value)

    def wrap_setter_column(self, fieldname):
        attr_name = anyblok_column_prefix + fieldname

        def setter_column(model_self, value):
            res = setattr(model_self, attr_name, value)
            for model_field, rfield in self.link_between_columns:
                self.apply_value_to(model_self, model_field, value, rfield)

            return res

        return setter_column


class InstrumentedAttribute_O2O(attributes.InstrumentedAttribute):

    def __init__(self, *args, **kwargs):
        self.relationship_field = kwargs.pop('relationship_field')
        super(InstrumentedAttribute_O2O, self).__init__(*args, **kwargs)

    def __set__(self, instance, value):
        call_super = False
        if value:
            for cname, fname in self.relationship_field.link_between_columns:
                if instance:
                    if getattr(instance, fname):
                        setattr(value, cname, getattr(instance, fname))
                    else:
                        call_super = True  # pragma: no cover
                else:
                    setattr(value, cname, instance)  # pragma: no cover

        else:
            call_super = True  # pragma: no cover

        if call_super:  # pragma: no cover
            super(InstrumentedAttribute_O2O, self).__set__(instance, value)


class One2One(Many2One):
    """ Define a relationship attribute on the model

    ::

        @register(Model)
        class TheModel:

            relationship = One2One(label="The relationship",
                                   model=Model.RemoteModel,
                                   remote_columns="The remote column",
                                   column_names="The column which have the "
                                               "foreigh key",
                                   nullable=False,
                                   backref="themodels")

    If the remote_columns are not define then, the system take the primary key
    of the remote model

    If the column doesn't exist, then the column will be create. Use the
    nullable option.
    If the name is not filled then the name is "'remote table'_'remote colum'"

    :param model: the remote model
    :param remote_columns: the column name on the remote model
    :param column_names: the column on the model which have the foreign key
    :param nullable: If the column_names is nullable
    :param backref: create the one2one link with this one2one
    """

    InstrumentedAttribute = InstrumentedAttribute_O2O

    def __init__(self, **kwargs):
        super(One2One, self).__init__(**kwargs)

        if 'backref' not in kwargs:
            raise FieldException("backref is a required argument")

        if 'one2many' in kwargs:
            raise FieldException("Unknow argmument 'one2many'")

        self.kwargs['info']['remote_name'] = self.kwargs['backref']

    def define_backref_properties(self, registry, namespace, properties):
        """ Add option uselist = False

        :param registry: the registry which load the relationship
        :param namespace: the name space of the model
        :param propertie: the properties known
        """
        self.backref_properties.update({'uselist': False})

    def apply_instrumentedlist(self, registry, namespace, fieldname):
        """ Add the InstrumentedList class to replace List class as result
        of the query

        :param registry: current registry
        """

    def get_relationship_cls(self):
        self.kwargs['relationship_field'] = self
        return RelationshipProperty


class Many2Many(RelationShip):
    """ Define a relationship attribute on the model

    ::

        @register(Model)
        class TheModel:

            relationship = Many2Many(label="The relationship",
                                     model=Model.RemoteModel,
                                     join_table="many2many table",
                                     remote_columns="The remote column",
                                     m2m_remote_columns="Name in many2many"
                                     local_columns="local primary key"
                                     m2m_local_columns="Name in many2many"
                                     many2many="themodels")

    if the join_table is not defined, then the table join is
        "join_'local table'_and_'remote table'"

    .. warning::

        The join_table must be filled when the declaration of the
        Many2Many is done in a Mixin

    If the remote_columns are not define then, the system take the primary key
    of the remote model

    if the local_columns are not define the take the primary key of the local
        model

    :param model: the remote model
    :param join_table: the many2many table to join local and remote models
    :param join_model: rich many2many where the join table come from a Model
    :param remote_columns: the column name on the remote model
    :param m2m_remote_columns: the column name to remote model in m2m table
    :param local_columns: the column on the model
    :param m2m_local_columns: the column name to local model in m2m table
    :param many2many: create the opposite many2many on the remote model
    """

    def __init__(self, **kwargs):
        super(Many2Many, self).__init__(**kwargs)

        self.join_table = self.kwargs.pop('join_table', None)
        self.join_model = self.kwargs.pop('join_model', None)
        if self.join_model:
            self.join_model = ModelAdapter(self.join_model)

        self.remote_columns = self.kwargs.pop('remote_columns', None)
        if self.remote_columns and not isinstance(self.remote_columns,
                                                  (list, tuple)):
            self.remote_columns = [self.remote_columns]

        self.m2m_remote_columns = self.kwargs.pop('m2m_remote_columns', None)
        if self.m2m_remote_columns and not isinstance(self.m2m_remote_columns,
                                                      (list, tuple)):
            self.m2m_remote_columns = [self.m2m_remote_columns]

        self.local_columns = self.kwargs.pop('local_columns', None)
        if self.local_columns and not isinstance(self.local_columns,
                                                 (list, tuple)):
            self.local_columns = [self.local_columns]

        self.m2m_local_columns = self.kwargs.pop('m2m_local_columns', None)
        if self.m2m_local_columns and not isinstance(self.m2m_local_columns,
                                                     (list, tuple)):
            self.m2m_local_columns = [self.m2m_local_columns]

        self.compute_join = self.kwargs.pop('compute_join', False)
        self.kwargs['backref'] = backref = self.kwargs.pop('many2many', None)
        self.kwargs['info']['remote_name'] = backref
        if isinstance(backref, (list, tuple)):
            self.kwargs['info']['remote_name'] = backref[0]
            self.backref_properties.update(**backref[1])

        self.schema = self.kwargs.pop('schema', None)

    def autodoc_get_properties(self):
        res = super(Many2Many, self).autodoc_get_properties()
        if self.join_table:
            res['join table'] = self.join_table

        if self.join_model:
            res['join model'] = self.join_model.model_name  # pragma: no cover

        if self.schema:
            res['schema'] = self.schema

        res['remote_columns'] = self.remote_columns
        res['m2m_remote_columns'] = self.m2m_remote_columns
        res['local_columns'] = self.local_columns
        res['m2m_local_columns'] = self.m2m_local_columns
        res['compute_join'] = self.compute_join
        return res

    def get_m2m_columns(self, registry, columns, m2m_columns, modelname,
                        suffix=""):
        if m2m_columns is None:
            m2m_columns = [
                x.get_fk_name(
                    registry, with_schema=False).replace('.', '_') + suffix
                for x in columns]
        elif self.join_model:
            m2m_columns_ = []
            first_step = registry.loaded_namespaces_first_step[
                self.join_model.model_name]
            for col in m2m_columns:
                if col not in first_step:
                    m2m_columns_.append(col)  # pragma: no cover
                elif isinstance(first_step[col], (Many2One, One2One)):
                    c = first_step[col]
                    remote_columns = c.get_remote_columns(registry)
                    m2m_columns_.extend([
                        x.attribute_name
                        for x in c.get_columns_names(
                            registry,
                            self.join_model.model_name,
                            col,
                            remote_columns
                        )
                    ])
                else:
                    m2m_columns_.append(col)

            m2m_columns = m2m_columns_

        if len(columns) != len(m2m_columns):
            raise FieldException((
                "The number of the column (%r) is not the same that the "
                "number m2m column (%r)") % (
                    columns, m2m_columns
                )
            )

        cols = []
        col_names = []
        ref_cols = []
        primaryjoin = []
        for i, column in enumerate(m2m_columns):
            sqltype = columns[i].native_type(registry)
            foreignkey = columns[i].get_fk_name(registry)
            completename = columns[i].get_complete_name(registry)
            cols.append(Column(column, sqltype, primary_key=True))
            col_names.append(column)
            ref_cols.append(foreignkey)
            primaryjoin.append(
                modelname + '.' + column + ' == ' + completename)

        primaryjoin = 'and_(' + ', '.join(primaryjoin) + ')'
        return cols, ForeignKeyConstraint(col_names, ref_cols), primaryjoin

    def get_local_and_remote_columns(self, registry):
        if not self.local_columns:
            local_columns = self.local_model.primary_keys(registry)
        else:
            local_columns = [ModelAttribute(self.local_model.model_name, x)
                             for x in self.local_columns]

        if not self.remote_columns:
            remote_columns = self.model.primary_keys(registry)
        else:
            remote_columns = [ModelAttribute(self.model.model_name, x)
                              for x in self.remote_columns]

        self.kwargs['info']['local_columns'] = [x.attribute_name
                                                for x in local_columns]
        self.kwargs['info']['remote_columns'] = [x.attribute_name
                                                 for x in remote_columns]

        return local_columns, remote_columns

    def get_join_table(self, registry, namespace, fieldname):
        """Get the join table name from join_table or join_model

        :param registry: the registry which load the relationship
        :param namespace: the name space of the model
        :param fieldname: fieldname of the relationship
        :rtype: name of the join table
        :exception: FieldException
        """
        join_table = self.join_table
        join_model_table = None
        if self.join_model:
            join_model_table = self.join_model.tablename(registry)

        if join_table and '.' in join_table:
            schema, table = join_table.split('.')
            schema = format_schema(schema, namespace)
            join_table = '%s.%s' % (schema, table)

        if join_table is None and join_model_table is None:
            join_table = ('join_%s_and_%s_for_%s' % (
                self.local_model.tablename(registry, with_schema=False),
                self.model.tablename(registry, with_schema=False),
                fieldname))[:63]

        elif join_table and join_model_table and join_table != join_model_table:
            raise FieldException(
                (
                    "The join_table %r and join_model %r is both declared, "
                    "on model %r and many2many %r, "
                    "but the both table name are different and we can not "
                    "determinate which is the good table's name"
                ) % (self.join_table, self.join_model.model_name,
                     namespace, fieldname)
            )

        return join_table or join_model_table

    def has_join_table_for_schema(self, registry, namespace, properties,
                                  join_table):
        has_join_table = False
        schema = None
        tables = registry.declarativebase.metadata.tables
        if '.' in join_table:
            has_join_table = join_table in tables
        elif self.join_model:
            has_join_table = join_table in tables
        elif self.schema:
            schema = format_schema(self.schema, namespace)
            has_join_table = self.schema + '.' + join_table in tables
        elif properties.get('__db_schema__'):
            schema = properties['__db_schema__']
            has_join_table = (
                properties['__db_schema__'] + '.' + join_table in tables)
        elif join_table in tables:
            has_join_table = True

        return has_join_table, schema

    def get_back_populate_relationship(self, registry, join_table):
        remote_model = self.model.model_name
        lnfs = registry.loaded_namespaces_first_step[remote_model]
        for fieldname in lnfs:
            field = lnfs[fieldname]
            if not isinstance(field, Many2Many):
                continue

            field.local_model = self.model
            remote_join_table = field.get_join_table(
                registry, remote_model, fieldname)
            if join_table == remote_join_table:
                return fieldname

        return None

    def set_overlaps_properties(self, registry, namespace):
        if not self.join_model:
            return

        lnfs = registry.loaded_namespaces_first_step[
            self.join_model.model_name]
        fieldnames = []
        for fieldname in lnfs:
            field = lnfs[fieldname]
            if not isinstance(field, Many2One):
                continue

            if field.model.model_name not in (
                namespace, self.model.model_name
            ):
                continue  # pragma: no cover

            fieldnames.append(anyblok_column_prefix + fieldname)

        if fieldnames:
            # M2O on join model to M2M fields
            self.kwargs['overlaps'] = ','.join(fieldnames)
            self.backref_properties['overlaps'] = ','.join(fieldnames)

    def get_sqlalchemy_mapping(self, registry, namespace, fieldname,
                               properties):
        """ Create the relationship

        :param registry: the registry which load the relationship
        :param namespace: the name space of the model
        :param fieldname: fieldname of the relationship
        :param properties: the properties known
        :rtype: Many2One relationship
        """
        self.model.check_model(registry)
        self.local_model = ModelRepr(namespace)
        local_columns, remote_columns = self.get_local_and_remote_columns(
            registry)
        join_table = self.get_join_table(registry, namespace, fieldname)
        has_join_table, schema = self.has_join_table_for_schema(
            registry, namespace, properties, join_table)
        if not has_join_table:
            modelname = ''.join(x.capitalize() for x in join_table.split('_'))
            remote_columns, remote_fk, secondaryjoin = self.get_m2m_columns(
                registry, remote_columns, self.m2m_remote_columns, modelname,
                suffix="right" if namespace == self.model.model_name else ""
            )
            local_columns, local_fk, primaryjoin = self.get_m2m_columns(
                registry, local_columns, self.m2m_local_columns, modelname,
                suffix="left" if namespace == self.model.model_name else ""
            )

            Node = Table(join_table, registry.declarativebase.metadata, *(
                local_columns + remote_columns + [local_fk, remote_fk]),
                schema=schema)

            if namespace == self.model.model_name:
                type(modelname, (registry.declarativebase,), {
                    '__table__': Node
                })
                self.kwargs['primaryjoin'] = primaryjoin
                self.kwargs['secondaryjoin'] = secondaryjoin

        elif namespace == self.model.model_name or self.compute_join:
            table = registry.declarativebase.metadata.tables[
                '%s.%s' % (schema, join_table) if schema else join_table]
            cls = get_class_by_table(registry.declarativebase, table)
            modelname = ModelRepr(cls.__registry_name__).modelname(registry)
            if (
                self.m2m_local_columns is None and
                self.m2m_remote_columns is None
            ):
                raise FieldException(  # pragma: no cover
                    "No 'm2m_local_columns' and 'm2m_remote_columns' "
                    "attribute filled for many2many "
                    "%r on model %r" % (fieldname, namespace))
            elif self.m2m_local_columns is None:
                raise FieldException(
                    "No 'm2m_local_columns' attribute filled for many2many "
                    "%r on model %r" % (fieldname, namespace))
            elif self.m2m_remote_columns is None:
                raise FieldException(
                    "No 'm2m_remote_columns' attribute filled for many2many"
                    " %r on model %r" % (fieldname, namespace))

            remote_columns, remote_fk, secondaryjoin = self.get_m2m_columns(
                registry, remote_columns, self.m2m_remote_columns,
                modelname,
                suffix="right" if namespace == self.model.model_name else ""
            )

            local_columns, local_fk, primaryjoin = self.get_m2m_columns(
                registry, local_columns, self.m2m_local_columns, modelname,
                suffix="left" if namespace == self.model.model_name else ""
            )
            self.kwargs['primaryjoin'] = primaryjoin
            self.kwargs['secondaryjoin'] = secondaryjoin

        self.set_overlaps_properties(registry, namespace)
        self.kwargs['secondary'] = (
            '%s.%s' % (schema, join_table) if schema else join_table)
        # definition of expiration
        if self.kwargs.get('backref'):
            self.init_expire_attributes(registry, namespace, fieldname)
            backref = self.kwargs['backref']
            if isinstance(backref, (tuple, list)):
                backref = backref[0]

            registry.expire_attributes[namespace][fieldname].add(
                ('x2m', fieldname, backref))
            model_name = self.model.model_name
            self.init_expire_attributes(registry, model_name, backref)
            registry.expire_attributes[model_name][backref].add(
                ('x2m', backref, fieldname))
        else:
            m2m = self.get_back_populate_relationship(registry, join_table)
            if m2m:
                self.kwargs['back_populates'] = m2m

        return super(Many2Many, self).get_sqlalchemy_mapping(
            registry, namespace, fieldname, properties)


class InstrumentedAttribute_O2M(attributes.InstrumentedAttribute):

    def __init__(self, *args, **kwargs):
        self.relationship_field = kwargs.pop('relationship_field')
        super(InstrumentedAttribute_O2M, self).__init__(*args, **kwargs)

    def __set__(self, instance, value):
        super(InstrumentedAttribute_O2M, self).__set__(instance, value)
        if instance:
            for cname, fname in self.relationship_field.link_between_columns:
                if value:
                    setattr(instance, cname, getattr(value, fname))
                else:
                    setattr(instance, cname, value)  # pragma: no cover


class One2Many(RelationShip):
    """ Define a relationship attribute on the model

    ::

        @register(Model)
        class TheModel:

            relationship = One2Many(label="The relationship",
                                    model=Model.RemoteModel,
                                    remote_columns="The remote column",
                                    primaryjoin="Join condition"
                                    many2one="themodel")

    If the primaryjoin is not filled then the join condition is
        "'local table'.'local promary key' == 'remote table'.'remote colum'"

    :param model: the remote model
    :param remote_columns: the column name on the remote model
    :param primaryjoin: the join condition between the remote column
    :param many2one: create the many2one link with this one2many
    """

    InstrumentedAttribute = InstrumentedAttribute_O2M

    def __init__(self, **kwargs):
        super(One2Many, self).__init__(**kwargs)

        self.remote_columns = None
        if 'remote_columns' in kwargs:
            remote_columns = self.kwargs.pop('remote_columns')
            if not isinstance(remote_columns, (list, tuple)):
                remote_columns = [remote_columns]

            self.remote_columns = [ModelAttribute(self.model.model_name, x)
                                   for x in remote_columns]

        if 'many2one' in kwargs:
            self.kwargs['backref'] = self.kwargs.pop('many2one')
            self.kwargs['info']['remote_name'] = self.kwargs['backref']

    def autodoc_get_properties(self):
        res = super(One2Many, self).autodoc_get_properties()
        res['remote_columns'] = self.remote_columns
        return res

    def add_expire_attributes(self, registry, namespace, fieldname):
        if self.kwargs.get('backref'):
            backref = self.kwargs['backref']
            if isinstance(backref, (list, tuple)):
                backref = backref[0]

            model_name = self.model.model_name
            for rname in self.remote_columns:
                self.init_expire_attributes(
                    registry, model_name, rname.attribute_name)
                _rname = rname.attribute_name
                registry.expire_attributes[model_name][_rname].add((backref,))
                registry.expire_attributes[model_name][_rname].add(
                    (backref, fieldname))

    def format_join_from_remote_columns(self, registry, namespace, fieldname):

        self.kwargs['info']['remote_columns'] = [x.attribute_name
                                                 for x in self.remote_columns]
        self.link_between_columns = [
            (x.attribute_name, x.get_fk_mapper(registry).attribute_name)
            for x in self.remote_columns
            if x.get_fk_mapper(registry)
        ]

        if 'primaryjoin' not in self.kwargs:
            pjs_ = {}
            for cname in self.remote_columns:
                remote = cname.get_complete_remote(registry)
                complete = cname.get_complete_name(registry)
                if remote in pjs_:
                    pjs_[remote].append(complete)
                else:
                    pjs_[remote] = [complete]

            pjs = []
            for k, v in pjs_.items():
                if len(v) == 1:
                    pjs.append("%s == %s" % (k, v[0]))
                else:
                    pj = 'or_(%s)' % ', '.join("%s == %s" % (k, y) for y in v)
                    logger.warning(
                        ("The One2Many %r on %r do a jointure on two identical "
                         "primary key : %r"), fieldname, namespace, pj)
                    pjs.append(pj)

            self.kwargs['primaryjoin'] = 'and_(' + ', '.join(pjs) + ')'

    def format_join_and_remote_columns(self, registry, namespace, fieldname):
        many2ones = self.model.many2one_for(registry, namespace)
        cmodel = self.model.model_name.replace('.', '')
        model = namespace.replace('.', '')
        pjs_ = {}
        self.link_between_columns = []
        self.kwargs['info']['remote_columns'] = []
        self.kwargs['info']['local_columns'] = []
        for m2o_name, many2one in many2ones:
            remote_columns = many2one.get_remote_columns(registry)
            for x in remote_columns:
                cname = m2o_name + '_' + x.attribute_name
                self.link_between_columns.append((cname, x.attribute_name))
                self.kwargs['info']['remote_columns'].append(cname)
                complete_name = cmodel + '.' + cname
                remote_name = model + '.' + x.attribute_name
                if remote_name in pjs_:
                    pjs_[remote_name].append(complete_name)
                else:
                    pjs_[remote_name] = [complete_name]

        if 'primaryjoin' not in self.kwargs:
            pjs = []
            for k, v in pjs_.items():
                if len(v) == 1:
                    pjs.append("%s == %s" % (k, v[0]))  # pragma: no cover
                else:
                    pj = 'or_(%s)' % ', '.join("%s == %s" % (k, y) for y in v)
                    logger.warning(
                        ("The One2Many %r on %r do a jointure on two identical "
                         "primary key : %r"), fieldname, namespace, pj)
                    pjs.append(pj)

            self.kwargs['primaryjoin'] = 'and_(' + ', '.join(pjs) + ')'

    def get_back_populate_relationship(self, registry, namespace):
        remote_model = self.model.model_name
        lnfs = registry.loaded_namespaces_first_step[remote_model]
        fieldnames = []
        for fieldname in lnfs:
            field = lnfs[fieldname]
            if not isinstance(field, Many2One):
                continue

            if field.model.model_name != namespace:
                continue  # pragma: no cover

            if field.kwargs.get('backref'):
                continue  # pragma: no cover

            fieldnames.append(anyblok_column_prefix + fieldname)

        return fieldnames

    def get_sqlalchemy_mapping(self, registry, namespace, fieldname,
                               properties):
        """ Create the relationship

        :param registry: the registry which load the relationship
        :param namespace: the name space of the model
        :param fieldname: fieldname of the relationship
        :param propertie: the properties known
        :rtype: Many2One relationship
        """
        self.model.check_model(registry)
        if not self.remote_columns:
            self.remote_columns = self.model.foreign_keys_for(
                registry, namespace)

        if self.remote_columns:
            self.format_join_from_remote_columns(registry, namespace, fieldname)
        else:
            self.format_join_and_remote_columns(registry, namespace, fieldname)

        self.kwargs['info']['local_columns'] = []
        for rcol in self.kwargs['info']['remote_columns']:
            col = ModelAttribute(
                self.model.model_name, rcol).get_fk_column(registry)
            if col:
                self.kwargs['info']['local_columns'].append(col)

        if not self.kwargs.get('backref'):
            m2o = self.get_back_populate_relationship(registry, namespace)
            if m2o:
                self.kwargs['overlaps'] = ','.join(m2o)

        self.add_expire_attributes(registry, namespace, fieldname)
        return super(One2Many, self).get_sqlalchemy_mapping(
            registry, namespace, fieldname, properties)

    def define_backref_properties(self, registry, namespace, properties):
        """ Add option in the backref if both model and remote model are the
        same, it is for the One2Many on the same model

        :param registry: the registry which load the relationship
        :param namespace: the name space of the model
        :param propertie: the properties known
        """
        if namespace == self.model.model_name:
            pks = ModelRepr(namespace).primary_keys(registry)
            self.backref_properties.update({'remote_side': [
                properties[anyblok_column_prefix + pk.attribute_name]
                for pk in pks]})

    def get_relationship_cls(self):
        self.kwargs['relationship_field'] = self
        return RelationshipProperty


def ordering_list(*args, **kwargs):
    fnct_args = args
    fnct_kwargs = kwargs

    def wrap(registry, *instrumented_list_bases, **properties):
        InstrumentedList = type(
            'InstrumentedList',
            (
                OrderingList.__mro__[0],
                *instrumented_list_bases,
                registry.InstrumentedList,
            ),
            properties
        )

        kw = _unsugar_count_from(**fnct_kwargs)
        return lambda: InstrumentedList(*fnct_args, **kw)

    wrap.is_an_anyblok_instrumented_list = True

    return wrap



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2019 Joachim Trouverie
#    Copyright (C) 2020 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#    Copyright (C) 2021 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from pkg_resources import iter_entry_points
from sqlalchemy.exc import IntegrityError, OperationalError
from alembic.migration import MigrationContext
from alembic.autogenerate import compare_metadata
from alembic.operations import Operations
from contextlib import contextmanager
from sqlalchemy import func, select, update, join, and_, text
from anyblok.config import Configuration
from sqlalchemy import inspect
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.ddl import CreateSchema, DropSchema
from .common import sgdb_in, return_list
from sqlalchemy.schema import (
    DDLElement, PrimaryKeyConstraint, CheckConstraint, UniqueConstraint)
from logging import getLogger

logger = getLogger(__name__)


MIGRATION_TYPE_PLUGINS_NAMESPACE = 'anyblok.migration_type.plugins'


class AlterSchema(DDLElement):
    def __init__(self, oldname, newname):
        self.oldname = oldname
        self.newname = newname


@compiles(CreateSchema, "mysql")
def compile_create_schema(element, compiler, **kw):
    schema = compiler.preparer.format_schema(element.element)
    return "CREATE SCHEMA %s CHARSET UTF8" % schema


@compiles(AlterSchema)
def compile_alter_schema(element, compiler, **kw):
    old_schema_name = compiler.preparer.format_schema(element.oldname)
    new_schema_name = compiler.preparer.format_schema(element.newname)
    return "ALTER SCHEMA %s RENAME TO %s" % (old_schema_name, new_schema_name)


@contextmanager
def cnx(migration):
    """ Context manager used by migration to get the connection """
    try:
        yield migration.conn
    except MigrationException:
        raise
    except Exception:  # pragma: no cover
        migration.conn.execute(text("rollback"))
        raise


class MigrationException(AttributeError):
    """ Simple Exception class for Migration """


class MigrationReport:
    """ Change report

    Get a new report::

        report = MigrationReport(migrationinstance, change_detected)

    """

    def ignore_migration_for(self, schema, table, default=None):
        if schema in self.ignore_migration_for_schema_from_configuration:
            return True

        if table in self.ignore_migration_for_table_from_configuration:
            return True

        return self.migration.ignore_migration_for.get(table, default)

    def raise_if_withoutautomigration(self):
        if self.migration.withoutautomigration:
            raise MigrationException("The metadata and the base structue are "
                                     "different, or this difference is "
                                     "forbidden in 'no auto migration' mode")

    def table_is_added(self, table):
        for action in self.actions:
            if action[0] == 'add_table' and action[1] is table:
                return True  # pragma: no cover

        return False

    def init_add_schema(self, diff):
        self.raise_if_withoutautomigration()
        _, schema = diff
        self.log_names.append('Add schema %s' % schema)

    def init_add_table(self, diff):
        self.raise_if_withoutautomigration()
        _, table = diff
        table_name = (
            '%s.%s' % (table.schema, table.name)
            if table.schema else table.name)
        self.log_names.append('Add table %s' % table_name)

    def init_add_column(self, diff):
        self.raise_if_withoutautomigration()
        _, schema, table, column = diff
        if self.ignore_migration_for(schema, table) is True:
            return True

        self.log_names.append('Add %s.%s' % (table, column.name))

    def can_remove_constraints(self, name):
        if name.startswith('anyblok_uq_'):
            return True

        if self.migration.reinit_constraints:
            return True

        if self.migration.reinit_all:
            return True

        return False

    def can_remove_fk_constraints(self, name):
        if name.startswith('anyblok_fk_'):
            return True

        if self.migration.reinit_constraints:
            return True

        if self.migration.reinit_all:
            return True

        return False

    def can_remove_check_constraints(self, name):
        if name.startswith('anyblok_ck_'):
            return True

        if self.migration.reinit_constraints:
            return True

        if self.migration.reinit_all:
            return True

        return False

    def init_remove_constraint(self, diff):
        _, constraint = diff
        if self.ignore_migration_for(constraint.table.schema,
                                     constraint.table.name) is True:
            return True

        self.log_names.append('Drop constraint %s on %s' % (
            constraint.name, constraint.table))

        if self.can_remove_constraints(constraint.name):
            self.raise_if_withoutautomigration()
        else:
            return True

    def can_remove_index(self, name):
        if name.startswith('anyblok_ix_'):
            return True

        if self.migration.reinit_indexes:
            return True

        if self.migration.reinit_all:
            return True

        return False

    def init_add_index(self, diff):
        self.raise_if_withoutautomigration()
        _, constraint = diff
        if self.ignore_migration_for(constraint.table.schema,
                                     constraint.table.name) is True:
            return True  # pragma: no cover

        columns = [x.name for x in constraint.columns]
        if self.table_is_added(constraint.table):
            return True  # pragma: no cover

        self.log_names.append('Add index constraint on %s (%s)' % (
            constraint.table.name, ', '.join(columns)))

    def init_remove_index(self, diff):
        _, index = diff

        if sgdb_in(self.migration.conn.engine, ['MySQL', 'MariaDB']):
            if index.table.schema in (
                'mysql',
                'performance_schema',
                'percona',
            ):
                return True

        if self.ignore_migration_for(index.table.schema,
                                     index.table.name) is True:
            return True

        self.log_names.append('Drop index %s on %s' % (index.name,
                                                       index.table))
        if self.can_remove_index(index.name):
            self.raise_if_withoutautomigration()
        else:
            return True

    def init_add_fk(self, diff):
        self.raise_if_withoutautomigration()
        _, fk = diff
        if self.ignore_migration_for(fk.table.schema, fk.table.name) is True:
            return True

        from_ = []
        to_ = []
        for column in fk.columns:
            if column.name in self.ignore_migration_for(
                fk.table.schema, fk.table.name, []
            ):
                return True

            for fk_ in column.foreign_keys:
                from_.append('%s.%s' % (fk.table.name, column.name))
                to_.append(fk_.target_fullname)

        self.log_names.append('Add Foreign keys on (%s) => (%s)' % (
            ', '.join(from_), ', '.join(to_)))

    def init_remove_fk(self, diff):
        _, fk = diff
        if self.ignore_migration_for(fk.table.schema, fk.table.name) is True:
            return True

        for column in fk.columns:
            if column.name in self.ignore_migration_for(
                fk.table.schema, fk.table.name, []
            ):
                return True

            for fk_ in column.foreign_keys:
                self.log_names.append('Drop Foreign keys on %s.%s => %s' % (
                    fk.table.name, column.name, fk_.target_fullname))

        if not self.can_remove_fk_constraints(fk.name):
            return True

        self.raise_if_withoutautomigration()

    def init_add_ck(self, diff):
        self.raise_if_withoutautomigration()
        _, table, ck = diff
        if self.ignore_migration_for(ck.table.schema, table) is True:
            return True

        if ck.table.schema:
            table = ck.table.schema + '.' + table

        self.log_names.append('Add check constraint %s on %s' % (
            ck.name, table))

    def init_remove_ck(self, diff):
        _, table, ck = diff
        if self.ignore_migration_for(ck['schema'], table) is True:
            return True

        if ck['schema']:
            table = ck['schema'] + '.' + table

        self.log_names.append('Drop check constraint %s on %s' % (
            ck['name'], table))

        if not self.can_remove_check_constraints(ck['name']):
            return True

        self.raise_if_withoutautomigration()

    def init_add_constraint(self, diff):
        self.raise_if_withoutautomigration()
        _, constraint = diff
        columns = []

        if self.ignore_migration_for(constraint.table.schema,
                                     constraint.table.name) is True:
            return True

        for column in constraint.columns:
            columns.append(column.name)
            if column.name in self.ignore_migration_for(constraint.table.schema,
                                                        constraint.table.name,
                                                        []):
                return True

        self.log_names.append('Add unique constraint on %s (%s)' % (
            constraint.table.name, ', '.join(columns)))

    def can_remove_column(self):
        if self.migration.reinit_columns:
            return True

        if self.migration.reinit_all:
            return True

        return False

    def init_remove_column(self, diff):
        column = diff[3]
        if self.ignore_migration_for(column.table.schema,
                                     column.table.name) is True:
            return True

        msg = "Drop Column %s.%s" % (column.table.name,
                                     column.name)

        if self.can_remove_column():
            self.log_names.append(msg)
            self.raise_if_withoutautomigration()
            return False

        fk_removed = []
        for fk in column.foreign_keys:
            if not self.can_remove_fk_constraints(fk.name):
                # only if fk is not removable. FK can come from
                # * DBA manager, it is the only raison to destroy it
                # * alembic, some constrainte change name during the remove
                if fk.name not in fk_removed:  # pragma: no cover
                    self.actions.append(('remove_fk', fk.constraint))
                    fk_removed.append(fk.name)

        if column.nullable is False:
            self.raise_if_withoutautomigration()
            msg += " (not null)"
            self.log_names.append(msg)
            self.actions.append(
                ('modify_nullable', column.table.schema, column.table.name,
                 column.name, {}, False, True))
            return True

        self.log_names.append(msg)
        return True

    def can_remove_table(self, schema):
        schemas = self.migration.metadata._schemas
        if schema and schema not in schemas:
            return False

        if self.migration.reinit_tables:
            return True

        if self.migration.reinit_all:
            return True

        return False

    def init_change_pk(self, diff):
        name, table, constraint = diff
        raise MigrationException(
            (
                "Change primary key constraint %s on %s: (%s). "
                "AnyBlok can't determine the good action to do "
                "for relation ship based on primary key who changed, "
                "You must make the migration by your self before."
            ) % (name, table, ', '.join([x.name for x in constraint.columns])))

    def init_remove_table(self, diff):
        table = diff[1]

        if sgdb_in(self.migration.conn.engine, ['MySQL', 'MariaDB']):
            if table.schema in (
                'mysql',
                'performance_schema',
                'percona',
            ):
                return True

        table_name = (
            '%s.%s' % (table.schema, table.name)
            if table.schema else table.name)
        self.log_names.append("Drop Table %s" % table_name)
        if self.can_remove_table(diff[1].schema):
            self.raise_if_withoutautomigration()
        else:
            return True

    def init_modify_type(self, diff):
        if self.ignore_migration_for(diff[1], diff[2]) is True:
            return True

        if diff[3] in self.ignore_migration_for(diff[1], diff[2], []):
            return True

        selected_plugin = self.get_plugin_for(diff[5], diff[6])
        if selected_plugin is not None:
            if not selected_plugin.need_to_modify_type():
                return True

        table = "%s.%s" % diff[1:3] if diff[1] else diff[2]
        self.log_names.append("Modify column type %s.%s : %s => %s" % (
            table, diff[3], diff[5], diff[6]))
        return False

    def init_modify_nullable(self, diff):
        if self.ignore_migration_for(diff[1], diff[2]) is True:
            return True

        if diff[3] in self.ignore_migration_for(diff[1], diff[2], []):
            return True

        table = "%s.%s" % diff[1:3] if diff[1] else diff[2]
        self.log_names.append("Modify column nullable %s.%s : %s => %s" % (
            table, diff[3], diff[5], diff[6]))
        return False

    def init_modify_server_default(self, diff):
        if self.ignore_migration_for(diff[1], diff[2]) is True:
            return True

        if diff[3] in self.ignore_migration_for(diff[1], diff[2], []):
            return True

        table = "%s.%s" % diff[1:3] if diff[1] else diff[2]
        self.log_names.append("Modify column default %s.%s : %s => %s" % (
            table, diff[3], diff[5], diff[6]))
        return False

    def init_plugins(self):
        """Get migration plugins from entry points"""

        def dialect_sort(plugin):
            """Sort plugins with dialect not None first"""
            return (plugin.dialect is None, plugin.dialect)

        plugins = sorted((
            entry_point.load()
            for entry_point
            in iter_entry_points(MIGRATION_TYPE_PLUGINS_NAMESPACE)
        ), key=dialect_sort)

        return plugins

    def get_plugin_for(self, oldvalue, newvalue):
        """search plugin by column types"""
        for plugin in self.plugins:
            if isinstance(plugin.dialect, (tuple, list)):
                dialects = plugin.dialect
            else:
                dialects = [plugin.dialect]

            if (
                issubclass(plugin, MigrationColumnTypePlugin) and
                isinstance(oldvalue, plugin.from_type) and
                isinstance(newvalue, plugin.to_type) and
                (
                    plugin.dialect is None or
                    sgdb_in(self.migration.conn.engine, dialects)
                )
            ):
                return plugin()

        return None

    def __init__(self, migration, diffs):
        """ Initializer

        :param migration: migration instance
        :param diffs: diff between the metadata and the database, come from
            change detection of alembic
        """
        self.migration = migration
        self.logs = []
        self.actions = []
        self.diffs = diffs
        self.log_names = []
        self.plugins = self.init_plugins()
        self.ignore_migration_for_table_from_configuration = [
            self.migration.loaded_namespaces[x].__tablename__
            for x in return_list(
                Configuration.get('ignore_migration_for_models')
            )
            if (
                x in self.migration.loaded_namespaces and
                self.migration.loaded_namespaces[x].is_sql
            )
        ]
        self.ignore_migration_for_schema_from_configuration = return_list(
            Configuration.get('ignore_migration_for_schemas'))

        mappers = {
            'add_schema': self.init_add_schema,
            'add_table': self.init_add_table,
            'add_column': self.init_add_column,
            'remove_constraint': self.init_remove_constraint,
            'add_index': self.init_add_index,
            'remove_index': self.init_remove_index,
            'add_fk': self.init_add_fk,
            'remove_fk': self.init_remove_fk,
            'add_ck': self.init_add_ck,
            'remove_ck': self.init_remove_ck,
            'add_constraint': self.init_add_constraint,
            'remove_column': self.init_remove_column,
            'remove_table': self.init_remove_table,
            'change_pk': self.init_change_pk,
            'modify_type': self.init_modify_type,
            'modify_nullable': self.init_modify_nullable,
            'modify_default': self.init_modify_server_default,
        }
        for diff in diffs:
            if isinstance(diff, list):
                self.raise_if_withoutautomigration()
                for change in diff:
                    _, _, table, column, _, _, _ = change
                    fnct = mappers.get(change[0])
                    if fnct:
                        if fnct(change):
                            continue
                    else:
                        logger.warning('Unknow diff: %r', change)

                    self.log_names.append('Alter %s.%s' % (table, column))
                    self.actions.append(change)
            else:
                fnct = mappers.get(diff[0])
                if fnct:
                    if fnct(diff):
                        continue
                else:
                    logger.warning('Unknow diff: %r', diff)

                self.actions.append(diff)

        for log_name in self.log_names:
            if log_name and not self.log_has(log_name):
                self.logs.append(log_name)

    def log_has(self, log):
        """ return True id the log is present

        .. warning:: this method is only used for the unittest

        :param log: log sentence expected
        """
        return log in self.logs

    def apply_change_add_schema(self, action):
        _, schema = action
        self.migration.schema().add(schema)

    def apply_change_add_table(self, action):
        _, table = action
        if table.schema:
            t = self.migration.schema(table.schema).table()
        else:
            t = self.migration.table()

        t.add(table.name, table=table)

    def get_migration_table(self, table):
        if table.schema:
            return self.migration.schema(table.schema).table(table.name)
        else:
            return self.migration.table(table.name)

    def apply_change_add_column(self, action):
        _, _, table, column = action
        t = self.get_migration_table(column.table)
        t.column().add(column)

    def apply_change_modify_nullable(self, action):
        _, schema, table, column, kwargs, oldvalue, newvalue = action
        if schema:
            t = self.migration.schema(schema).table(table)
        else:
            t = self.migration.table(table)

        t.column(column).alter(
            nullable=newvalue, existing_nullable=oldvalue, **kwargs)

    def apply_change_modify_type(self, action):
        _, schema, table, column, kwargs, oldvalue, newvalue = action
        if schema:
            t = self.migration.schema(schema).table(table)
        else:
            t = self.migration.table(table)

        selected_plugin = self.get_plugin_for(oldvalue, newvalue)
        if selected_plugin is not None:
            selected_plugin.apply(t.column(column), **kwargs)
        else:
            t.column(column).alter(
                type_=newvalue, existing_type=oldvalue, **kwargs)

    def apply_change_modify_default(self, action):
        _, schema, table, column, kwargs, oldvalue, newvalue = action
        if schema:
            t = self.migration.schema(schema).table(table)  # pragma: no cover
        else:
            t = self.migration.table(table)

        t.column(column).alter(
            server_default=newvalue, existing_server_default=oldvalue,
            **kwargs)

    def apply_change_remove_constraint(self, action):
        _, constraint = action
        if constraint.__class__ is UniqueConstraint:
            table = self.get_migration_table(constraint.table)
            table.unique(name=constraint.name).drop()

    def apply_change_remove_index(self, action):
        _, index = action
        if not index.unique:
            table = self.get_migration_table(index.table)
            table.index(name=index.name).drop()

    def apply_change_add_fk(self, action):
        _, fk = action
        t = self.get_migration_table(fk.table)
        from_ = []
        to_ = []
        for column in fk.columns:
            for fk_ in column.foreign_keys:
                from_.append(column.name)
                to_.append(fk_.column)

        t.foreign_key(fk.name).add(from_, to_)

    def apply_change_add_ck(self, action):
        _, table, ck = action
        t = self.get_migration_table(ck.table)
        t.check(ck.name).add(ck.sqltext)

    def apply_change_remove_fk(self, action):
        _, fk = action
        t = self.get_migration_table(fk.table)
        t.foreign_key(fk.name).drop()

    def apply_change_remove_ck(self, action):
        _, table, ck = action
        if ck['schema']:
            t = self.migration.schema(ck['schema']).table(table)
        else:
            t = self.migration.table(table)

        t.foreign_key(ck['name']).drop()

    def apply_change_add_constraint(self, action):
        _, constraint = action
        table = self.get_migration_table(constraint.table)
        table.unique(name=constraint.name).add(*constraint.columns)

    def apply_change_add_index(self, action):
        _, constraint = action
        table = self.get_migration_table(constraint.table)
        table.index().add(*constraint.columns, name=constraint.name)

    def apply_remove_table(self, action):
        table = self.get_migration_table(action[1])
        table.drop()

    def apply_remove_column(self, action):
        table = self.get_migration_table(action[3].table)
        table.column(action[3].name).drop()

    def apply_change(self):
        """ Apply the migration

        this method parses the detected change and calls the Migration
        system to apply the change with the api of Declarations
        """
        for log in self.logs:
            logger.debug(log)

        mappers = {
            'add_schema': self.apply_change_add_schema,
            'add_table': self.apply_change_add_table,
            'add_column': self.apply_change_add_column,
            'modify_nullable': self.apply_change_modify_nullable,
            'modify_type': self.apply_change_modify_type,
            'modify_default': self.apply_change_modify_default,
            'add_index': self.apply_change_add_index,
            'add_fk': self.apply_change_add_fk,
            'add_ck': self.apply_change_add_ck,
            'add_constraint': self.apply_change_add_constraint,
            'remove_constraint': self.apply_change_remove_constraint,
            'remove_index': self.apply_change_remove_index,
            'remove_fk': self.apply_change_remove_fk,
            'remove_ck': self.apply_change_remove_ck,
            'remove_table': self.apply_remove_table,
            'remove_column': self.apply_remove_column,
        }
        for action in self.actions:
            fnct = mappers.get(action[0])
            if fnct:
                fnct(action)


class MigrationConstraintForeignKey:
    """ Used to apply a migration on a foreign key

    You can add::

        table.column('my column').foreign_key().add(Blok.name)

    Or drop::

        table.column('my column').foreign_key().drop()

    """
    def __init__(self, table, name):
        self.table = table
        self.name = name

    def add(self, local_columns, remote_columns, **kwargs):
        """ Add a new foreign key

        :param remote_field: The column of the remote model
        :rtype: MigrationConstraintForeignKey instance
        """
        remote_columns = [
            x.property.columns[0] if hasattr(x, 'property') else x
            for x in remote_columns]

        remote_table = set(x.table.name for x in remote_columns)
        if len(remote_table) != 1:
            raise MigrationException(  # pragma: no cover
                "Remote column must have the same table "
                "(%s)" % ', '.join(remote_table))

        remote_table = remote_table.pop()
        remote_columns_names = [x.name for x in remote_columns]
        self.table.migration.operation.create_foreign_key(
            self.name, self.table.name, remote_table,
            local_columns, remote_columns_names,
            source_schema=self.table.schema,
            referent_schema=remote_columns[0].table.schema,
            **kwargs)

        return self

    def drop(self):
        """ Drop the foreign key """
        self.table.migration.operation.drop_constraint(
            self.name, self.table.name, type_='foreignkey',
            schema=self.table.schema)
        return self


class MigrationColumnTypePlugin:
    """Meta class for column migration type plugin

    Must be exposed as entry point in namespace 'anyblok.migration_type.plugins'

    :param to_type: Column type value (sqlalchemy.types) as used in Model
                    classes in source code
    :param from_type: Column type value (sqlalchemy.types) as required to
                      communicate with the DBMS
    :param dialect: DB dialect (list of strings or string)

    Example::

    class BooleanToTinyIntMySQL(MigrationColumnTypePlugin):

        to_type = sqlalchemy.types.Boolean
        from_type = sqlalchemy.types.TINYINT
        dialect = ['MySQL', 'MariaDB']

        def need_to_modify_type(self):
            return False

        def apply(self, column, **kwargs):
            '''Boolean are TINYINT in MySQL DataBases'''
            # do nothing
            pass
    """

    to_type = None
    from_type = None
    dialect = None

    def apply(self, column, **kwargs):
        """Apply column migration, this method MUST be overriden in plugins
        subclass
        """
        raise NotImplementedError()  # pragma: no cover

    def need_to_modify_type(self, column, **kwargs):
        """If False the type won't be modified"""
        return True  # pragma: no cover


class MigrationColumn:
    """ get or add a column

    Add a new column::

        table.column().add(Sqlachemy column)

    Get a column::

        c = table.column('My column name')

    Alter the column::

        c.alter(new_column_name='Another column name')

    Drop the column::

        c.drop()
    """
    def __init__(self, table, name):
        self.table = table
        self.name = name
        self.info = {}

        if name is not None:
            op = self.table.migration.operation
            with cnx(self.table.migration) as conn:
                columns = op.impl.dialect.get_columns(
                    conn, self.table.name, schema=table.schema)

            for c in columns:
                if c['name'] == name:
                    self.info.update(c)

            if not self.info:
                raise MigrationException(
                    "No column %r found on %r" % (name, self.table.name))

    def apply_default_value(self, column):
        if column.default:
            execute = self.table.migration.conn.execute
            val = column.default.arg
            table = self.table.migration.metadata.tables[self.table.name]
            table.append_column(column)
            cname = getattr(table.c, column.name)
            if column.default.is_callable:  # pragma: no cover
                Table = self.table.migration.metadata.tables['system_model']
                Column = self.table.migration.metadata.tables['system_column']
                j1 = join(Table, Column, Table.c.name == Column.c.model)
                query = select([Column.c.name]).select_from(j1)
                query = query.where(Column.c.primary_key.is_(True))
                query = query.where(Table.c.table == self.table.name)
                columns = [x[0] for x in execute(query).fetchall()]

                query = select([func.count()]).select_from(table)
                query = query.where(cname.is_(None))
                nb_row = self.table.migration.conn.execute(query).fetchone()[0]
                for offset in range(nb_row):
                    query = select(columns).select_from(table)
                    query = query.where(cname.is_(None)).limit(1)
                    res = execute(query).fetchone()
                    where = and_(
                        *[getattr(table.c, x) == res[x] for x in columns])
                    query = update(table).where(where).values(
                        {cname: val(None)})
                    execute(query)

            else:
                query = update(table).where(cname.is_(None)).values(
                    {cname: val})
                execute(query)

    def add(self, column):
        """ Add a new column

        The column is added in two phases, the last phase is only for the
        the nullable, if nullable can not be applied, a warning is logged

        :param column: sqlalchemy column
        :rtype: MigrationColumn instance
        """
        migration = self.table.migration
        nullable = column.nullable
        if not nullable:
            column.nullable = True

        # check the table exist
        table = (
            "%s.%s" % (self.table.schema, self.table.name)
            if self.table.schema
            else self.table.name)

        table_ = migration.metadata.tables[table]

        if sgdb_in(self.table.migration.conn.engine, ['MsSQL']):
            column.table = table_

        migration.operation.impl.add_column(self.table.name, column,
                                            schema=self.table.schema)
        self.apply_default_value(column)

        if not nullable:
            c = MigrationColumn(self.table, column.name)
            c.alter(nullable=False)

        return MigrationColumn(self.table, column.name)

    def alter(self, **kwargs):
        """ Alter an existing column

        Alter the column in two phases, because the nullable column has not
        locked the migration

        .. warning::
            See Alembic alter_column, the existing_* param are used for some
            dialect like mysql, is importante to filled them for these dialect

        :param new_column_name: New name for the column
        :param type_: New sqlalchemy type
        :param server_default: The default value in database server
        :param nullable: New nullable value
        :param comment: New comment value
        :rtype: MigrationColumn instance
        """
        vals = {}
        name = self.name
        if 'existing_server_default' in kwargs:
            esd = kwargs['existing_server_default']
            if esd:
                vals['existing_server_default'] = esd.arg
            else:
                vals['existing_server_default'] = esd
        else:
            vals['existing_server_default'] = (
                self.server_default
                if 'server_default' not in kwargs
                else None)
        vals.update({
            'existing_type': kwargs.get(
                'existing_type',
                self.type if 'type_' not in kwargs else None),
            'existing_autoincrement': (
                None
                if not sgdb_in(self.table.migration.conn.engine,
                               ['MySQL', 'MariaDB'])
                else kwargs.get(
                    'existing_autoincrement',
                    self.autoincrement
                    if 'autoincrement' not in kwargs else None)
            ),
            'existing_comment': kwargs.get(
                'existing_comment',
                self.comment if 'comment' not in kwargs else None)
        })
        vals.update({
            k: kwargs[k]
            for k in ('autoincrement', 'server_default', 'type_')
            if k in kwargs})

        if 'name' in kwargs:
            vals['new_column_name'] = kwargs['name']
            name = kwargs['name']

        if vals:
            self.table.migration.operation.alter_column(
                self.table.name, self.name,
                schema=self.table.schema, **vals)

        if 'nullable' in kwargs:
            nullable = kwargs['nullable']
            vals['existing_nullable'] = (
                self.nullable if 'nullable' in kwargs else None)
            savepoint = '%s_not_null' % name
            try:
                self.table.migration.savepoint(savepoint)
                self.table.migration.operation.alter_column(
                    self.table.name, self.name, nullable=nullable,
                    schema=self.table.schema, **vals)
                self.table.migration.release_savepoint(savepoint)
            except (IntegrityError, OperationalError) as e:
                self.table.migration.rollback_savepoint(savepoint)
                logger.warning(str(e))

        return MigrationColumn(self.table, name)

    def drop(self):
        """ Drop the column """
        self.table.migration.operation.drop_column(
                self.table.name, self.name, schema=self.table.schema)

    @property
    def nullable(self):
        """ Use for unittest return if the column is nullable """
        return self.info.get('nullable', None)

    @property
    def type(self):
        """ Use for unittest: return the column type """
        return self.info.get('type', None)

    @property
    def server_default(self):
        """ Use for unittest: return the default database value """
        sdefault = self.info.get('default', None)
        if sgdb_in(self.table.migration.conn.engine, ['MySQL', 'MariaDB']):
            if sdefault:
                if not isinstance(sdefault, str):
                    return sdefault.arg  # pragma: no cover
                elif sdefault is None:
                    return None  # pragma: no cover
                else:
                    return text(sdefault)

        return sdefault

    @property
    def comment(self):
        """ Use for unittest: return the default database value """
        return self.info.get('comment', None)

    @property
    def autoincrement(self):
        """ Use for unittest: return the default database value """
        table_name = (
            "%s.%s" % (self.table.schema, self.table.name)
            if self.table.schema else self.table.name)
        table = self.table.migration.metadata.tables[table_name]
        primary_keys = [x.name for x in table.primary_key.columns]
        if self.name in primary_keys:
            return False  # pragma: no cover

        return self.info.get('autoincrement', None)


class MigrationConstraintCheck:
    """ Used for the Check constraint

    Add a new constraint::

        table('My table name').check().add('check_my_column', 'mycolumn > 5')

    Get and drop the constraint::

        table('My table name').check('check_my_column').drop()

    """
    def __init__(self, table, name):
        self.table = table
        self.name = name
        # TODO dialect not have method to check if constraint exist

    def add(self, condition):
        """ Add the constraint

        :param condition: constraint to apply
        :rtype: MigrationConstraintCheck instance
        """
        self.table.migration.operation.create_check_constraint(
            self.name, self.table.name, condition,
            schema=self.table.schema)

        return self

    def drop(self):
        """ Drop the constraint """
        self.table.migration.operation.drop_constraint(
            self.name, self.table.name, type_='check',
            schema=self.table.schema)


class MigrationConstraintUnique:
    """ Used for the Unique constraint

    Add a new constraint::

        table('My table name').unique('constraint name').add('col1', 'col2')

    Get and drop the constraint::

        table('My table name').unique('constraint name').drop()

    Let AnyBlok to define the name of the constraint::

        table('My table name').unique(None).add('col1', 'col2')


    """
    def __init__(self, table, name):
        self.table = table
        self.name = name

    def add(self, *columns):
        """ Add the constraint

        :param *columns: list of SQLalchemy column
        :rtype: MigrationConstraintUnique instance
        :exception: MigrationException
        """
        if not columns:
            raise MigrationException(  # pragma: no cover
                """To add an unique constraint you """
                """must define one or more columns""")

        columns_name = [x.name for x in columns]
        savepoint = 'uq_%s' % (self.name or '')
        try:
            self.table.migration.savepoint(savepoint)
            self.table.migration.operation.create_unique_constraint(
                self.name, self.table.name, columns_name,
                schema=self.table.schema)
            self.table.migration.release_savepoint(savepoint)
        except (IntegrityError, OperationalError) as e:
            self.table.migration.rollback_savepoint(savepoint)

            logger.warning(
                "Error during the add of new unique constraint %r "
                "on table %r and columns %r : %r " % (self.name,
                                                      self.table.name,
                                                      columns_name, str(e)))

        return self

    def drop(self):
        """ Drop the constraint """
        self.table.migration.operation.drop_constraint(
            self.name, self.table.name, type_='unique',
            schema=self.table.schema)


class MigrationConstraintPrimaryKey:
    """ Used for the primary key constraint

    Add a new constraint::

        table('My table name').primarykey().add('col1', 'col2')

    Get and drop the constraint::

        table('My table name').primarykey('col1', 'col2').drop()
    """

    def __init__(self, table):
        self.table = table
        self.name = self.format_name()

    def format_name(self, *columns):
        return 'anyblok_pk_%s' % self.table.name

    def add(self, *columns):
        """ Add the constraint

        :param *columns: list of SQLalchemy column
        :rtype: MigrationConstraintPrimaryKey instance
        :exception: MigrationException
        """
        if not columns:
            raise MigrationException(  # pragma: no cover
                """To add a primary key constraint """
                """you must define one or more columns""")

        if sgdb_in(self.table.migration.conn.engine, ['MsSQL']):
            for column in columns:
                if column.nullable:
                    column.alter(nullable=False)

        columns_name = [x.name for x in columns]
        self.table.migration.operation.create_primary_key(
            self.name, self.table.name, columns_name)
        return self

    def drop(self):
        """ Drop the constraint """
        self.table.migration.operation.drop_constraint(
            self.name, self.table.name, type_='primary')
        return self


class MigrationIndex:
    """ Used for the index constraint

    Add a new constraint::

        table('My table name').index().add('col1', 'col2')

    Get and drop the constraint::

        table('My table name').index('col1', 'col2').drop()

    """

    def __init__(self, table, *columns, **kwargs):
        self.table = table
        if 'name' in kwargs:
            self.name = kwargs['name']
        else:
            self.name = self.format_name(*columns)
        self.exist = False

        if self.name is not None:
            op = self.table.migration.operation
            with cnx(self.table.migration) as conn:
                indexes = op.impl.dialect.get_indexes(
                    conn, self.table.name, schema=self.table.schema)

            for i in indexes:
                if i['name'] == self.name:
                    self.exist = True

            if not self.exist:
                raise MigrationException(  # pragma: no cover
                    "No index %r found on %r" % (self.name, self.table.name))

    def format_name(self, *columns):
        if columns:
            cols = [x.name for x in columns]
            cols.sort()
            cols = '_'.join(cols)
            return 'idx_%s_on_%s' % (cols, self.table.name)

        return None

    def add(self, *columns, **kwargs):
        """ Add the constraint

        :param *columns: list of SQLalchemy column
        :param **kwargs: other attribute fir l __init__
        :rtype: MigrationIndex instance
        :exception: MigrationException
        """
        if not columns:
            raise MigrationException(  # pragma: no cover
                "To add an index you must define one or more columns")

        index_name = kwargs.get('name', self.format_name(*columns))
        columns_name = [x.name for x in columns]
        self.table.migration.operation.create_index(
            index_name, self.table.name, columns_name,
            schema=self.table.schema)

        return MigrationIndex(self.table, *columns, **kwargs)

    def drop(self):
        """ Drop the constraint """
        self.table.migration.operation.drop_index(
            self.name, table_name=self.table.name,
            schema=self.table.schema)


class MigrationTable:
    """ Use to manipulate tables

    Add a table::

        table().add('New table')

    Get an existing table::

        t = table('My table name')

    Alter the table::

        t.alter(name='Another table name')

    Drop the table::

        t.drop()
    """

    def __init__(self, migration, name, schema=None):
        self.name = name
        self.migration = migration
        self.schema = schema

        if name is not None:
            with cnx(self.migration) as conn:
                has_table = migration.operation.impl.dialect.has_table
                if not has_table(conn, name, schema=schema):
                    raise MigrationException("No table %r found" % name)

    def add(self, name, table=None):
        """ Add a new table

        :param name: name of the table
        :param table: an existing instance of the table to create
        :rtype: MigrationTable instance
        """
        if table is not None:
            if table.schema != self.schema:
                raise MigrationException(  # pragma: no cover
                    "The schema of the table (%r.%r) and the MigrationTable %r"
                    "instance are not the same" % (
                        table.schema, table.name, self.schema))

            self.migration.metadata.create_all(
                bind=self.migration.conn,
                tables=[table])
        else:
            self.migration.operation.create_table(name, schema=self.schema)

        return MigrationTable(self.migration, name, self.schema)

    def column(self, name=None):
        """ Get Column

        :param name: Column name
        :rtype: MigrationColumn instance
        """
        return MigrationColumn(self, name)

    def drop(self):
        """ Drop the table """
        self.migration.operation.drop_table(self.name, schema=self.schema)

    def index(self, *columns, **kwargs):
        """ Get index

        :param *columns: List of the column's name
        :rtype: MigrationIndex instance
        """
        return MigrationIndex(self, *columns, **kwargs)

    def unique(self, name):
        """ Get unique

        :param name: str name of the unique constraint
        :rtype: MigrationConstraintUnique instance
        """
        return MigrationConstraintUnique(self, name)

    def check(self, name=None):
        """ Get check

        :param name: str name of the check constraint
        :rtype: MigrationConstraintCheck instance
        """
        return MigrationConstraintCheck(self, name)

    def primarykey(self):
        """ Get primary key

        :rtype: MigrationConstraintPrimaryKey instance
        """
        return MigrationConstraintPrimaryKey(self)

    def alter(self, **kwargs):
        """ Atler the current table

        :param name: New table name
        :rtype: MigrationTable instance
        :exception: MigrationException
        """
        if 'name' not in kwargs:
            raise MigrationException(  # pragma: no cover
                "Table can only alter name")

        name = kwargs['name']
        self.migration.operation.rename_table(
            self.name, name, schema=self.schema)
        return MigrationTable(self.migration, name, schema=self.schema)

    def foreign_key(self, name):
        """ Get a foreign key

        :rtype: MigrationConstraintForeignKey instance
        """
        return MigrationConstraintForeignKey(self, name)


class MigrationSchema:
    """ Use to manipulate tables

    Add a Schema::

        schema().add('New schema')

    Get an existing schema::

        s = schema('My table schema')

    Alter the schema::

        s.alter(name='Another schema name')

    Drop the schema::

        s.drop()
    """

    def __init__(self, migration, name):
        self.name = name
        self.migration = migration

        if name is not None:
            if not self.has_schema():
                raise MigrationException("No schema %r found" % self.name)

    def has_schema(self):
        with cnx(self.migration) as conn:
            if sgdb_in(conn.engine, ['MySQL', 'MariaDB', 'MsSQL']):
                query = """
                    SELECT count(*)
                    FROM INFORMATION_SCHEMA.SCHEMATA
                    WHERE SCHEMA_name='%s'
                """ % self.name
                return conn.execute(text(query)).fetchone()[0]
            else:
                return self.migration.operation.impl.dialect.has_schema(
                    conn, self.name)

    def add(self, name):
        """ Add a new schema

        :param name: name of the schema
        :rtype: MigrationSchema instance
        """
        with cnx(self.migration) as conn:
            conn.execute(CreateSchema(name))

        return MigrationSchema(self.migration, name)

    def table(self, name=None):
        """ Get a table

        :rtype: MigrationTable instance
        """
        return MigrationTable(self.migration, name, schema=self.name)

    def alter(self, name=None):
        """ Atler the current table

        :param name: New schema name
        :rtype: MigrationSchema instance
        :exception: MigrationException
        """
        with cnx(self.migration) as conn:
            conn.execute(AlterSchema(self.name, name))

        return MigrationSchema(self.migration, name)

    def drop(self, cascade=False):
        """ Drop the schema """
        with cnx(self.migration) as conn:
            conn.execute(DropSchema(self.name, cascade=cascade))


class Migration:
    """ Migration Main entry

    This class allows to manipulate all the migration class::

        migration = Migration(Session(), Base.Metadata)
        t = migration.table('My table name')
        c = t.column('My column name from t')
    """

    def __init__(self, registry):
        self.withoutautomigration = registry.withoutautomigration
        self.conn = registry.connection()
        self.loaded_namespaces = registry.loaded_namespaces
        self.loaded_views = registry.loaded_views
        self.metadata = registry.declarativebase.metadata
        self.ddl_compiler = self.conn.dialect.ddl_compiler(
            self.conn.dialect, None)
        self.ignore_migration_for = registry.ignore_migration_for

        opts = {
            'include_schemas': True,
            'compare_server_default': True,
            'render_item': self.render_item,
            'compare_type': self.compare_type,
        }
        self.context = MigrationContext.configure(self.conn, opts=opts)
        self.operation = Operations(self.context)
        self.reinit_all = Configuration.get('reinit_all', False)
        self.reinit_tables = Configuration.get('reinit_tables', False)
        self.reinit_columns = Configuration.get('reinit_columns', False)
        self.reinit_indexes = Configuration.get('reinit_indexes', False)
        self.reinit_constraints = Configuration.get(
            'reinit_constraints', False)

    def table(self, name=None, schema=None):
        """ Get a table

        :param name: default None, name of the table
        :param schema: default None, name of the schema

        :rtype: MigrationTable instance
        """
        return MigrationTable(self, name=name, schema=schema)

    def schema(self, name=None):
        """ Get a table

        :rtype: MigrationSchema instance
        """
        return MigrationSchema(self, name)

    def auto_upgrade_database(self, schema_only=False):
        """ Upgrade the database automaticly """
        report = self.detect_changed(schema_only=schema_only)
        report.apply_change()

    def detect_changed(self, schema_only=False):
        """ Detect the difference between the metadata and the database

        :rtype: MigrationReport instance
        """
        inspector = inspect(self.conn)
        if schema_only:
            diff = self.detect_added_new_schema(inspector)
        else:
            diff = compare_metadata(self.context, self.metadata)
            diff.extend(
                self.detect_undetected_constraint_from_alembic(inspector))

        return MigrationReport(self, diff)

    def detect_added_new_schema(self, inspector):
        diff = []
        schemas = self.metadata._schemas
        reflected_schemas = set(inspector.get_schema_names())
        added_schemas = schemas - reflected_schemas
        for schema in added_schemas:
            diff.append(('add_schema', schema))

        return diff

    def detect_undetected_constraint_from_alembic(self, inspector):
        diff = []
        diff.extend(self.detect_check_constraint_changed(inspector))
        diff.extend(self.detect_pk_constraint_changed(inspector))
        return diff

    def check_constraint_is_same(self, reflected_constraint, constraint):
        """the goal is to detect if contrainst changed when the name is long

        SQLAlchemy trunkated the name if function of database type (
        postgres 63 characters)

        this method check if the truncated name is the same that no truncated
        name and if the constraint text is the same: return True else False
        """
        truncated_name = self.ddl_compiler.preparer.format_constraint(
            constraint)
        if truncated_name == reflected_constraint['name']:
            return True

        return False  # pragma: no cover

    def detect_check_constraint_changed(self, inspector):
        if sgdb_in(self.conn.engine, ['MySQL', 'MariaDB', 'MsSQL']):
            # MySQL don t return the reflected constraint
            return []

        diff = []
        schemas = list(self.metadata._schemas)
        schemas.append(None)
        for schema in schemas:
            for table in inspector.get_table_names(schema=schema):
                table_ = "%s.%s" % (schema, table) if schema else table
                if table_ not in self.metadata.tables:
                    continue

                reflected_constraints = {
                    ck['name']: ck
                    for ck in inspector.get_check_constraints(
                        table, schema=schema)
                }
                constraints = {
                    ck.name: ck
                    for ck in self.metadata.tables[table_].constraints
                    if isinstance(ck, CheckConstraint)
                    if ck.name != '_unnamed_'
                }
                todrop = set(reflected_constraints.keys()) - set(
                    constraints.keys())
                toadd = set(constraints.keys()) - set(
                    reflected_constraints.keys())

                # check a constraint have not been truncated
                todrop_ = todrop.copy()
                for x in todrop_:
                    for y in toadd:
                        if self.check_constraint_is_same(
                            reflected_constraints[x], constraints[y]
                        ):
                            toadd.remove(y)
                            todrop.remove(x)
                            break

                for ck in todrop:
                    ck_ = reflected_constraints[ck]
                    ck_['schema'] = schema
                    diff.append(('remove_ck', table, ck_))

                for ck in toadd:
                    diff.append(('add_ck', table, constraints[ck]))

        return diff

    def detect_pk_constraint_changed(self, inspector):
        diff = []
        schemas = list(self.metadata._schemas)
        schemas.append(None)
        for schema in schemas:
            for table in inspector.get_table_names(schema=schema):
                table_ = "%s.%s" % (schema, table) if schema else table
                if table_ not in self.metadata.tables:
                    continue

                reflected_constraint = inspector.get_pk_constraint(
                    table, schema=schema)
                constraint = [
                    pk
                    for pk in self.metadata.tables[table_].constraints
                    if isinstance(pk, PrimaryKeyConstraint)
                ][0]
                reflected_columns = set(
                    reflected_constraint['constrained_columns'])
                columns = set(x.name for x in constraint.columns)
                if columns != reflected_columns:
                    diff.append(('change_pk', table, constraint))

        return diff

    def savepoint(self, name=None):
        """ Add a savepoint

        :param name: name of the save point
        :rtype: return the name of the save point
        """
        if sgdb_in(self.conn.engine, ['MySQL', 'MariaDB']):
            logger.warning(
                "Try to create a SAVEPOINT, but %r don't have this "
                "functionality" % self.conn.engine.dialect)
            return

        return self.conn._savepoint_impl(name=name)

    def rollback_savepoint(self, name):
        """ Rollback to the savepoint

        :param name: name of the savepoint
        """
        if sgdb_in(self.conn.engine, ['MySQL', 'MariaDB']):
            logger.warning(
                "Try to ROLLBACK TO SAVEPOINT, but %r don't have this "
                "functionality" % self.conn.engine.dialect)
            return

        self.conn._rollback_to_savepoint_impl(name)

    def release_savepoint(self, name):
        """ Release the save point

        :param name: name of the savepoint
        """
        if sgdb_in(self.conn.engine, ['MySQL', 'MariaDB']):
            logger.warning(
                "Try to RELEASE SAVEPOINT, but %r don't have this "
                "functionality" % self.conn.engine.dialect)
            return

        self.conn._release_savepoint_impl(name)

    def render_item(self, type_, obj, autogen_context):
        logger.debug("%r, %r, %r" % (type_, obj, autogen_context))
        return False

    def compare_type(self, context, inspected_column,
                     metadata_column, inspected_type, metadata_type):
        if hasattr(metadata_type, 'compare_type'):
            return metadata_type.compare_type(  # pragma: no cover
                inspected_type)

        return None



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Denis VIVIÈS <dvivies@geoblink.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest


@pytest.mark.usefixtures('rollback_registry')
class TestCoreSqlBase:

    def test_insert(self, rollback_registry):
        registry = rollback_registry
        Blok = registry.System.Blok
        Blok.insert(name='OneBlok', state='undefined', version='0.0.0')
        blok = Blok.query().filter(Blok.name == 'OneBlok').first()
        assert blok.state == 'undefined'

    def test_multi_insert(self, rollback_registry):
        registry = rollback_registry
        Blok = registry.System.Blok
        Blok.multi_insert(
            dict(name='OneBlok', state='undefined', version='0.0.0'),
            dict(name='TwoBlok', state='undefined', version='0.0.0'),
            dict(name='ThreeBlok', state='undefined', version='0.0.0'))
        states = Blok.query('state').filter(Blok.name.in_(['OneBlok',
                                                           'TwoBlok',
                                                           'ThreeBlok'])).all()
        states = [x[0] for x in states]
        assert states == ['undefined', 'undefined', 'undefined']

    def test_from_primary_key(self, rollback_registry):
        registry = rollback_registry
        Model = registry.System.Model
        model = Model.query().first()
        model2 = Model.from_primary_keys(name=model.name)
        assert model == model2

    def test_from_primary_keys(self, rollback_registry):
        registry = rollback_registry
        Column = registry.System.Column
        column = Column.query().first()
        column2 = Column.from_primary_keys(model=column.model,
                                           name=column.name)
        assert column == column2

    def test_get_primary_key(self, rollback_registry):
        registry = rollback_registry
        assert registry.System.Model.get_primary_keys() == ['name']

    def test_get_primary_keys(self, rollback_registry):
        registry = rollback_registry
        pks = registry.System.Column.get_primary_keys()
        model_in_pks = 'model' in pks
        name_in_pks = 'name' in pks
        assert model_in_pks is True
        assert name_in_pks is True

    def test_to_primary_key(self, rollback_registry):
        registry = rollback_registry
        model = registry.System.Model.query().first()
        assert model.to_primary_keys() == dict(name=model.name)

    def test_to_primary_keys(self, rollback_registry):
        registry = rollback_registry
        column = registry.System.Column.query().first()
        assert column.to_primary_keys() == {
            'model': column.model, 'name': column.name}

    def test_fields_description(self, rollback_registry):
        registry = rollback_registry
        Model = registry.System.Model
        res = {'description': {'id': 'description',
                               'label': 'Description',
                               'model': None,
                               'nullable': True,
                               'primary_key': False,
                               'type': 'Function'},
               'is_sql_model': {'id': 'is_sql_model',
                                'label': 'Is a SQL model',
                                'model': None,
                                'nullable': True,
                                'primary_key': False,
                                'type': 'Boolean'},
               'name': {'id': 'name',
                        'label': 'Name',
                        'model': None,
                        'nullable': False,
                        'primary_key': True,
                        'type': 'String'},
               'table': {'id': 'table',
                         'label': 'Table',
                         'model': None,
                         'nullable': True,
                         'primary_key': False,
                         'type': 'String'},
               'schema': {'id': 'schema',
                          'label': 'Schema',
                          'model': None,
                          'nullable': True,
                          'primary_key': False,
                          'type': 'String'}}
        assert Model.fields_description() == res

    def test_fields_description_limited_field(self, rollback_registry):
        registry = rollback_registry
        Model = registry.System.Model
        res = {'table': {'id': 'table',
                         'label': 'Table',
                         'model': None,
                         'nullable': True,
                         'primary_key': False,
                         'type': 'String'}}
        assert Model.fields_description(fields=['table']) == res

    def test_fields_description_cache(self, rollback_registry):
        registry = rollback_registry
        Model = registry.System.Model
        Column = registry.System.Column
        res = {'table': {'id': 'table',
                         'label': 'Table',
                         'model': None,
                         'nullable': True,
                         'primary_key': False,
                         'type': 'String'}}
        assert Model.fields_description(fields=['table']) == res
        column = Column.from_primary_keys(model='Model.System.Model',
                                          name='table')
        column.label = 'Test'
        assert Model.fields_description(fields=['table']) == res
        Model.fire('Update Model', 'Model.System.Model')
        assert Model.fields_description(fields=['table']) != res

    def test_to_dict(self, rollback_registry):
        registry = rollback_registry
        M = registry.System.Model
        model = M.query().first()
        assert model.to_dict() == {
            'name': model.name,
            'table': model.table,
            'schema': model.schema,
            'is_sql_model': model.is_sql_model,
            'description': model.description,
        }

    def test_to_dict_on_some_columns(self, rollback_registry):
        registry = rollback_registry
        M = registry.System.Model
        model = M.query().first()
        assert model.to_dict('name', 'table') == {
            'name': model.name,
            'table': model.table,
        }

    def test_select_sql_statement_with_column(self, rollback_registry):
        registry = rollback_registry
        Blok = registry.System.Blok
        assert Blok.execute_sql_statement(
            Blok.select_sql_statement('name').where(
                Blok.name == 'anyblok-core').limit(1)
        ).one() == ('anyblok-core',)

    def test_from_multi_primary_keys(self, rollback_registry):
        registry = rollback_registry
        Blok = registry.System.Blok
        blok_names = Blok.from_multi_primary_keys(
            dict(name='anyblok-core'),
            dict(name='anyblok-test'),
        ).name
        assert 'anyblok-core' in blok_names
        assert 'anyblok-test' in blok_names

    def test_from_multi_primary_keys_empty(self, rollback_registry):
        registry = rollback_registry
        Blok = registry.System.Blok
        assert Blok.from_multi_primary_keys() == []

    def test_get_hybrid_property_columns(self, rollback_registry):
        registry = rollback_registry
        Column = registry.System.Column
        columns = Column.get_hybrid_property_columns()
        for x in ['name', 'model', 'autoincrement', 'foreign_key',
                  'primary_key', 'unique', 'nullable', 'remote_model']:
            assert x in columns



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2020 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.testing import sgdb_in


@pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']), reason='ISSUE #89')
@pytest.mark.usefixtures('rollback_registry')
class TestSystemSequence:

    def test_nextval_without_prefix_without_suffix(self, rollback_registry):
        registry = rollback_registry
        Sequence = registry.System.Sequence
        seq = Sequence.insert(code='test.sequence')
        assert seq.current is None
        assert seq.nextval() == '1'
        assert seq.nextval() == '2'
        assert seq.nextval() == '3'
        assert seq.current == 3

    def test_nextval_with_start_value(self, rollback_registry):
        registry = rollback_registry
        Sequence = registry.System.Sequence
        seq = Sequence.insert(code='test.sequence', start=10)
        assert seq.current is None
        assert seq.nextval() == '10'
        assert seq.nextval() == '11'
        assert seq.nextval() == '12'
        assert seq.current == 12

    def test_nextval_without_prefix_without_suffix_two_time(
        self, rollback_registry
    ):
        registry = rollback_registry
        Sequence = registry.System.Sequence
        Sequence.insert(code='test.sequence')
        seq = Sequence.insert(code='test.sequence')
        current = seq.current
        assert current is None
        assert seq.nextval() == '1'
        assert seq.nextval() == '2'
        assert seq.nextval() == '3'

    def test_nextval_without_prefix_with_suffix(self, rollback_registry):
        registry = rollback_registry
        Sequence = registry.System.Sequence
        seq = Sequence.insert(code='test.sequence', formater="{seq}_suffix")
        current = seq.current
        assert current is None
        assert seq.nextval() == '%d_suffix' % 1
        assert seq.nextval() == '%d_suffix' % 2
        assert seq.nextval() == '%d_suffix' % 3

    def test_nextval_with_prefix_without_suffix(self, rollback_registry):
        registry = rollback_registry
        Sequence = registry.System.Sequence
        seq = Sequence.insert(code='test.sequence', formater='prefix_{seq}')
        current = seq.current
        assert current is None
        assert seq.nextval() == 'prefix_%d' % 1
        assert seq.nextval() == 'prefix_%d' % 2
        assert seq.nextval() == 'prefix_%d' % 3

    def test_nextval_with_prefix_with_suffix(self, rollback_registry):
        registry = rollback_registry
        Sequence = registry.System.Sequence
        seq = Sequence.insert(code='test.sequence',
                              formater='prefix_{seq}_suffix')
        current = seq.current
        assert current is None
        assert seq.nextval() == 'prefix_%d_suffix' % 1
        assert seq.nextval() == 'prefix_%d_suffix' % 2
        assert seq.nextval() == 'prefix_%d_suffix' % 3

    def test_nextval_by_attribute(self, rollback_registry):
        registry = rollback_registry
        Sequence = registry.System.Sequence
        seq = Sequence.insert(code='test.sequence')
        current = seq.current
        assert current is None
        assert Sequence.nextvalBy(code=seq.code) == '1'
        assert Sequence.nextvalBy(code=seq.code) == '2'
        assert Sequence.nextvalBy(code=seq.code) == '3'

    def test_sequence_with_gap_rollback(self, rollback_registry):
        registry = rollback_registry
        Sequence = registry.System.Sequence
        seq = Sequence.insert(code='test.sequence', no_gap=False)
        current = seq.current
        assert current is None
        assert Sequence.nextvalBy(code=seq.code) == '1'
        registry.commit()
        assert Sequence.nextvalBy(code=seq.code) == '2'
        registry.rollback()
        assert Sequence.nextvalBy(code=seq.code) == '3'

    def test_sequence_with_gap_rollback_and_start(self, rollback_registry):
        registry = rollback_registry
        Sequence = registry.System.Sequence
        seq = Sequence.insert(
            code='test.sequence.start', no_gap=False, start=10)
        assert seq.current is None
        assert Sequence.nextvalBy(code=seq.code) == '10'
        registry.commit()
        assert Sequence.nextvalBy(code=seq.code) == '11'
        registry.rollback()
        assert Sequence.nextvalBy(code=seq.code) == '12'
        assert seq.current == 12

    def test_sequence_with_no_gap_rollback(self, rollback_registry):
        registry = rollback_registry
        Sequence = registry.System.Sequence
        seq = Sequence.insert(code='test.sequence.nogap', no_gap=True)
        current = seq.current
        assert current is None
        assert Sequence.nextvalBy(code=seq.code) == '1'
        registry.commit()
        assert Sequence.nextvalBy(code=seq.code) == '2'
        registry.rollback()
        assert Sequence.nextvalBy(code=seq.code) == '2'
